# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-12-19 15:17
from __future__ import unicode_literals

from django.db import migrations, models
from django.db.models import Count
from tqdm import tqdm


def remove_duplicate_legal_reasons(apps, purpose_slug, source_object_content_type, source_object_id):
    LegalReason = apps.get_model('gdpr', 'LegalReason')
    duplicate_legal_reason_qs = LegalReason.objects.filter(
        purpose_slug=purpose_slug,
        source_object_content_type=source_object_content_type,
        source_object_id=source_object_id
    )

    if duplicate_legal_reason_qs.filter(is_active=True).count() > 0:
        duplicate_legal_reason_qs.filter(is_active=False).delete()

    latest_legal_reason = duplicate_legal_reason_qs.latest('expires_at')
    duplicate_legal_reason_qs.exclude(pk=latest_legal_reason.pk).delete()


def check_uniqueness_and_keep_latest_active_legal_reason(apps, schema_editor):
    LegalReason = apps.get_model('gdpr', 'LegalReason')
    check_qs = LegalReason.objects.values('purpose_slug', 'source_object_content_type', 'source_object_id').annotate(
        lr_count=Count('purpose_slug')).filter(lr_count__gt=1).order_by('-lr_count').distinct()

    for legal_reason in tqdm(check_qs.all()):
        remove_duplicate_legal_reasons(
            apps, legal_reason['purpose_slug'], legal_reason['source_object_content_type'],
            legal_reason['source_object_id']
        )


def remove_duplicate_legal_reasons_relatives(apps, legal_reason, object_content_type, object_id):
    LegalReasonRelatedObject = apps.get_model('gdpr', 'LegalReasonRelatedObject')
    duplicates_qs = LegalReasonRelatedObject.objects.filter(
        legal_reason=legal_reason,
        object_content_type=object_content_type,
        object_id=object_id
    )
    latest_legal_reason_related_object = duplicates_qs.latest('created_at')
    duplicates_qs.exclude(pk=latest_legal_reason_related_object.pk).delete()


def check_uniqueness_and_keep_latest_active_legal_reason_related_object(apps, schema_editor):
    LegalReasonRelatedObject = apps.get_model('gdpr', 'LegalReasonRelatedObject')
    check_qs = LegalReasonRelatedObject.objects.values('legal_reason', 'object_content_type', 'object_id').annotate(
        lrro_count=Count('legal_reason')).filter(lrro_count__gt=1).order_by('-lrro_count').distinct()

    for legal_reason_related_object in tqdm(check_qs.all(), ncols=100):
        remove_duplicate_legal_reasons_relatives(apps, legal_reason_related_object['legal_reason'],
                                                 legal_reason_related_object['object_content_type'],
                                                 legal_reason_related_object['object_id']
                                                 )


class Migration(migrations.Migration):
    dependencies = [
        ('gdpr', '0002_auto_20180509_1518'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]
    operations = [
        migrations.AlterField(
            model_name='legalreason',
            name='purpose_slug',
            field=models.CharField(choices=[], db_index=True,
                                   max_length=100, verbose_name='purpose'),
        ),
        migrations.AlterField(
            model_name='legalreason',
            name='source_object_id',
            field=models.TextField(verbose_name='source object ID', db_index=True),
        ),
        migrations.AlterField(
            model_name='legalreasonrelatedobject',
            name='object_id',
            field=models.TextField(verbose_name='related object ID', db_index=True),
        ),
        migrations.RunPython(check_uniqueness_and_keep_latest_active_legal_reason),
        migrations.RunPython(check_uniqueness_and_keep_latest_active_legal_reason_related_object),
    ]



from typing import TYPE_CHECKING, Any, Dict, KeysView, List, Optional, Tuple, Type, Union

from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model, Q
from django.apps import apps
from django.db.models.utils import make_model_tuple

from gdpr.enums import LegalReasonState
from gdpr.fields import Fields
from gdpr.loading import anonymizer_register, purpose_register

if TYPE_CHECKING:
    from gdpr.models import LegalReason
    from gdpr.anonymizers import ModelAnonymizer

FieldList = Union[List[str], Tuple, KeysView[str]]  # List, tuple or return of dict keys() method.
FieldMatrix = Union[str, Tuple[Any, ...]]
RelatedMatrix = Dict[str, FieldMatrix]


class PurposeMetaclass(type):

    def __new__(mcs, name, bases, attrs):
        from gdpr.loading import purpose_register

        new_class = super().__new__(mcs, name, bases, attrs)
        if hasattr(new_class, 'slug') and new_class.slug:
            if new_class.slug in purpose_register:
                raise ImproperlyConfigured('More anonymization purposes with slug {}'.format(new_class.slug))

            purpose_register.register(new_class.slug, new_class)

        def set_source_model_class(model):
            new_class.source_model_class = model

        if isinstance(new_class.source_model_class, str):
            apps.lazy_model_operation(set_source_model_class, make_model_tuple(new_class.source_model_class))

        return new_class

    def __str__(self):
        return str(self.name)


class AbstractPurpose(metaclass=PurposeMetaclass):
    """

    :param anonymize_legal_reason_related_object_only: If True anonymize only related objects which have links which
    have LegalReasonRelatedObject records.
    """

    name: str
    slug: str
    fields: Union[str, Tuple[Any, ...], None] = None
    expiration_timedelta: Any
    anonymize_legal_reason_related_objects_only: bool = False  # @TODO: Add support
    source_model_class: Optional[Union[str, Type[Model]]] = None

    def can_anonymize_obj(self, obj: Model, fields: FieldMatrix):
        return len(fields) != 0

    def get_parsed_fields(self, model: Type[Model]) -> Fields:
        return Fields(self.fields or (), model)

    def deanonymize_obj(self, obj: Model, fields: Optional[FieldMatrix] = None):
        fields = fields or self.fields or ()
        if len(fields) == 0:
            # If there are no fields to deanonymize do nothing.
            return
        obj_model = obj.__class__
        anonymizer: "ModelAnonymizer" = anonymizer_register[obj_model]()
        anonymizer.deanonymize_obj(obj, fields)

    def anonymize_obj(self, obj: Model, legal_reason: Optional["LegalReason"] = None,
                      fields: Optional[FieldMatrix] = None):
        fields = fields or self.fields or ()
        if not self.can_anonymize_obj(obj, fields):
            # If there are no fields to anonymize do nothing.
            return
        from gdpr.models import LegalReason  # noqa

        obj_model = obj.__class__
        anonymizer: "ModelAnonymizer" = anonymizer_register[obj_model]()

        # MultiLegalReason
        other_legal_reasons = LegalReason.objects.filter_source_instance(obj).filter(state=LegalReasonState.ACTIVE)
        if legal_reason:
            other_legal_reasons = other_legal_reasons.filter(~Q(pk=legal_reason.pk))
        if other_legal_reasons.count() == 0:
            anonymizer.anonymize_obj(obj, legal_reason, self, fields)
            return

        from gdpr.loading import purpose_register

        parsed_fields = self.get_parsed_fields(obj_model)

        # Transform legal_reasons to fields
        for allowed_fields in [purpose_register[slug]().get_parsed_fields(obj_model) for slug in
                               set([i.purpose_slug for i in other_legal_reasons])]:
            parsed_fields -= allowed_fields
            if len(parsed_fields) == 0:
                # If there are no fields to anonymize do nothing.
                return

        anonymizer.anonymize_obj(obj, legal_reason, self, parsed_fields)


purposes_map = purpose_register  # Backwards compatibility



from gdpr import anonymizers
from gdpr.anonymizers.local import cs
from tests.models import (
    Account, Address, ContactForm, Customer, CustomerRegistration, Email, Note, Payment, Avatar, ChildE
)


class CustomerRegistrationAnonymizer(anonymizers.ModelAnonymizer):
    email_address = anonymizers.EmailFieldAnonymizer()

    class Meta:
        model = Customer


class CustomerAnonymizer(anonymizers.ModelAnonymizer):
    first_name = anonymizers.MD5TextFieldAnonymizer()
    last_name = anonymizers.MD5TextFieldAnonymizer()
    primary_email_address = anonymizers.EmailFieldAnonymizer()

    full_name = anonymizers.CharFieldAnonymizer()
    birth_date = anonymizers.DateFieldAnonymizer()
    personal_id = cs.CzechPersonalIDSmartFieldAnonymizer()
    phone_number = cs.CzechPhoneNumberFieldAnonymizer()
    facebook_id = anonymizers.CharFieldAnonymizer()
    last_login_ip = anonymizers.IPAddressFieldAnonymizer()
    last_registration = CustomerRegistrationAnonymizer()
    other_registrations = CustomerRegistrationAnonymizer()

    notes = anonymizers.ReverseGenericRelationAnonymizer('tests', 'Note')

    def get_encryption_key(self, obj: Customer):
        return (f"{(obj.first_name or '').strip()}::{(obj.last_name or '').strip()}::"
                f"{(obj.primary_email_address or '').strip()}")

    class Meta:
        model = Customer


class EmailAnonymizer(anonymizers.ModelAnonymizer):
    email = anonymizers.EmailFieldAnonymizer()

    class Meta:
        model = Email


class AddressAnonymizer(anonymizers.ModelAnonymizer):
    street = anonymizers.CharFieldAnonymizer()

    class Meta:
        model = Address


class AccountAnonymizer(anonymizers.ModelAnonymizer):
    number = cs.CzechAccountNumberFieldAnonymizer(use_smart_method=True)
    IBAN = cs.CzechIBANSmartFieldAnonymizer()
    owner = anonymizers.CharFieldAnonymizer()

    class Meta:
        model = Account


class PaymentAnonymizer(anonymizers.ModelAnonymizer):
    value = anonymizers.DecimalFieldAnonymizer()
    date = anonymizers.DateFieldAnonymizer()

    class Meta:
        model = Payment


class ContactFormAnonymizer(anonymizers.ModelAnonymizer):
    email = anonymizers.EmailFieldAnonymizer()
    full_name = anonymizers.CharFieldAnonymizer()

    class Meta:
        model = ContactForm
        reversible_anonymization = False


class NoteAnonymizer(anonymizers.ModelAnonymizer):
    note = anonymizers.CharFieldAnonymizer()
    contact_form = anonymizers.GenericRelationAnonymizer('tests', 'ContactForm')

    class Meta:
        model = Note


class AvatarAnonymizer(anonymizers.ModelAnonymizer):
    image = anonymizers.ReplaceFileFieldAnonymizer()

    class Meta:
        model = Avatar


class ChildEAnonymizer(anonymizers.ModelAnonymizer):
    name = anonymizers.CharFieldAnonymizer()
    first_name = anonymizers.CharFieldAnonymizer()
    last_name = anonymizers.CharFieldAnonymizer()
    birth_date = anonymizers.DateFieldAnonymizer()
    note = anonymizers.CharFieldAnonymizer()

    class Meta:
        model = ChildE
        anonymize_reversion = True



import json
from decimal import Decimal

from django.test import TestCase
from django.utils import timezone

from gdpr.anonymizers import (
    CharFieldAnonymizer, DateFieldAnonymizer, DecimalFieldAnonymizer, EmailFieldAnonymizer, IPAddressFieldAnonymizer,
    StaticValueFieldAnonymizer
)
from gdpr.anonymizers.fields import (
    DateTimeFieldAnonymizer, FunctionFieldAnonymizer, IntegerFieldAnonymizer, JSONFieldAnonymizer,
    SiteIDUsernameFieldAnonymizer
)
from germanium.tools import assert_dict_equal, assert_equal, assert_list_equal, assert_not_equal


class TestCharField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = CharFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_char_field(self):
        name = 'John CENA'
        out = self.field.get_encrypted_value(name, self.encryption_key)

        assert_not_equal(out, name)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, name)

    def test_char_field_transliteration(self):
        name = 'François'
        fixed_name = 'Francois'
        field = CharFieldAnonymizer(transliterate=True)
        out = field.get_encrypted_value(name, self.encryption_key)

        assert_not_equal(out, name)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, fixed_name)

    def test_char_field_transliteration_full_czech(self):
        text = 'Příliš žluťoučký kůň úpěl ďábelské ódy'
        fixed_text = 'Prilis zlutoucky kun upel dabelske ody'
        field = CharFieldAnonymizer(transliterate=True)
        out = field.get_encrypted_value(text, self.encryption_key)

        assert_not_equal(out, text)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, fixed_text)


class TestEmailField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = EmailFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_email_field(self):
        email = 'foo@bar.com'
        out = self.field.get_encrypted_value(email, self.encryption_key)

        assert_not_equal(out, email)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, email)


class TestSiteIDUsernameFieldAnonymizer(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = SiteIDUsernameFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_just_email(self):
        email = 'foo@bar.com'
        out = self.field.get_encrypted_value(email, self.encryption_key)

        assert_not_equal(out, email)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, email)

    def test_normal(self):
        email = '1:foo@bar.com'
        out = self.field.get_encrypted_value(email, self.encryption_key)

        assert_not_equal(out, email)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, email)


class TestDateField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = DateFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_date_field(self):
        date = timezone.now()
        out = self.field.get_encrypted_value(date, self.encryption_key)

        assert_not_equal(out, date)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, date)


class TestDateTimeField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = DateTimeFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_date_field(self):
        date = timezone.now()
        out = self.field.get_encrypted_value(date, self.encryption_key)

        assert_not_equal(out, date)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, date)


class TestDecimalField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = DecimalFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_decimal_field_positive(self):
        decimal = Decimal('3.14159265358979')
        out = self.field.get_encrypted_value(decimal, self.encryption_key)

        assert_not_equal(out, decimal)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, decimal)

    def test_decimal_field_negative(self):
        decimal = Decimal('-3.14159265358979')
        out = self.field.get_encrypted_value(decimal, self.encryption_key)

        assert_not_equal(out, decimal)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, decimal)


class TestIntegerField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = IntegerFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_integer_field_positive(self):
        number = 42
        out = self.field.get_encrypted_value(number, self.encryption_key)

        assert_not_equal(out, number)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, number)

    def test_integer_field_negative(self):
        number = -42
        out = self.field.get_encrypted_value(number, self.encryption_key)

        assert_not_equal(out, number)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, number)


class TestIPAddressField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = IPAddressFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_ip_addr_v_4_field(self):
        ip = '127.0.0.1'
        out = self.field.get_encrypted_value(ip, self.encryption_key)

        assert_not_equal(out, ip)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, ip)

    def test_ip_addr_v_6_field(self):
        ip = '::1'
        out = self.field.get_encrypted_value(ip, self.encryption_key)

        assert_not_equal(out, ip)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, ip)


class TestStaticValueAnonymizer(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = StaticValueFieldAnonymizer('U_SHALL_NOT_PASS')
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_static_field(self):
        text = 'ORANGE'
        out = self.field.get_encrypted_value(text, self.encryption_key)

        assert_not_equal(out, text)


class TestFunctionField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_one_way_function_field(self):
        number = 5
        self.field = FunctionFieldAnonymizer(lambda x, key: x ** 2)
        out = self.field.get_encrypted_value(number, self.encryption_key)

        assert_not_equal(out, number)

    def test_two_way_function_field(self):
        number = 5
        self.field = FunctionFieldAnonymizer(lambda a, x, key: x + a.get_numeric_encryption_key(key),
                                             lambda a, x, key: x - a.get_numeric_encryption_key(key))
        self.field.max_anonymization_range = 100
        out = self.field.get_encrypted_value(number, self.encryption_key)

        assert_not_equal(out, number)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, number)


class TestJSONFieldAnonymizer(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = JSONFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_str_value(self):
        text = 'John CENA'
        out = self.field.anonymize_json_value(text, self.encryption_key)

        assert_not_equal(out, text)
        out_decrypt = self.field.anonymize_json_value(out, self.encryption_key, False)

        assert_equal(out_decrypt, text)

    def test_none_value(self):
        value = None
        out = self.field.anonymize_json_value(value, self.encryption_key)

        assert_equal(out, value)
        out_decrypt = self.field.anonymize_json_value(out, self.encryption_key, False)

        assert_equal(out_decrypt, value)

    def test_int_value(self):
        value = 158
        out = self.field.anonymize_json_value(value, self.encryption_key)

        assert_not_equal(out, value)
        out_decrypt = self.field.anonymize_json_value(out, self.encryption_key, False)

        assert_equal(out_decrypt, value)

    def test_int_value_overflow(self):
        value = 9
        out = self.field.anonymize_json_value(value, self.encryption_key)

        assert_not_equal(out, value)
        out_decrypt = self.field.anonymize_json_value(out, self.encryption_key, False)

        assert_equal(out_decrypt, value)

    def test_dict(self):
        json_dict = {
            'breed': 'labrador',
            'owner': {
                'name': 'Bob',
                'other_pets': [{'name': 'Fishy'}]
            },
            'age': 5,
            'height': 9.5,
            'is_brown': True,
            'none_field': None
        }

        out = self.field.anonymize_json_value(json_dict, self.encryption_key)

        assert_not_equal(out, json_dict)

        out_decrypt = self.field.anonymize_json_value(out, self.encryption_key, False)

        assert_dict_equal(json_dict, out_decrypt)

    def test_dict_str(self):
        json_dict = {
            'breed': 'labrador',
            'owner': {
                'name': 'Bob',
                'other_pets': [{'name': 'Fishy'}]
            },
            'age': 5,
            'height': 9.5,
            'is_brown': True,
            'none_field': None
        }

        out = json.loads(self.field.get_encrypted_value(json.dumps(json_dict), self.encryption_key))

        assert_not_equal(out, json_dict)

        out_decrypt = json.loads(self.field.get_decrypted_value(json.dumps(out), self.encryption_key))

        assert_dict_equal(json_dict, out_decrypt)

    def test_list(self):
        json_list = ['banana', 'oranges', 5, 3.14, False, None, {'name': 'Bob'}]

        out = self.field.anonymize_json_value(json_list, self.encryption_key)

        assert_not_equal(out, json_list)

        out_decrypt = self.field.anonymize_json_value(out, self.encryption_key, False)

        assert_list_equal(json_list, out_decrypt)

    def test_list_str(self):
        json_list = ['banana', 'oranges', 5, 3.14, False, None, {'name': 'Bob'}]

        out = json.loads(self.field.get_encrypted_value(json.dumps(json_list), self.encryption_key))

        assert_not_equal(out, json_list)

        out_decrypt = json.loads(self.field.get_decrypted_value(json.dumps(out), self.encryption_key))

        assert_list_equal(json_list, out_decrypt)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-05-09 13:18
from __future__ import unicode_literals

import datetime

from django.db import migrations, models
from django.utils.timezone import utc


class Migration(migrations.Migration):
    dependencies = [
        ('gdpr', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='legalreasonrelatedobject',
            old_name='content_type',
            new_name='object_content_type',
        ),
        migrations.AddField(
            model_name='legalreason',
            name='issued_at',
            field=models.DateTimeField(default=datetime.datetime(2018, 5, 9, 13, 18, 7, 317147, tzinfo=utc),
                                       verbose_name='issued at'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='legalreason',
            name='purpose_slug',
            field=models.CharField(choices=[], max_length=100, verbose_name='purpose'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2020-02-17 13:08
from __future__ import unicode_literals

from django.db import migrations
import enumfields.fields
import gdpr.enums


def set_legal_reason_state(apps, schema_editor):
    LegalReason = apps.get_model('gdpr', 'LegalReason')
    LegalReason.objects.filter(is_active=True).update(state=gdpr.enums.LegalReasonState.ACTIVE)
    LegalReason.objects.filter(is_active=False).update(state=gdpr.enums.LegalReasonState.EXPIRED)


class Migration(migrations.Migration):

    dependencies = [
        ('gdpr', '0007_migration'),
    ]

    operations = [
        migrations.RunPython(set_legal_reason_state),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2019-03-05 09:23
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0005_avatar'),
    ]

    operations = [
        migrations.RenameField(
            model_name='avatar',
            old_name='custormer',
            new_name='customer',
        ),
    ]



from typing import Optional

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db.models import Model

from .base import RelationAnonymizer


class ReverseGenericRelationAnonymizer(RelationAnonymizer):
    """Defines relation for anonymizer to cope with GenericForeignKey."""

    app_name: str
    model_name: str
    content_type_field: str
    id_field: str

    def __init__(self, app_name: str, model_name: Optional[str] = None, content_type_field: str = 'content_type',
                 id_field: str = 'object_id'):
        """

        :param app_name: The name of the app or `<app_name>.<model_name>`
        :param model_name: The name of the model with GenericRelation
        :param content_type_field: The name of the FK to ContentType Model
        :param id_field: The id of the related model
        """
        if model_name is None:
            self.app_name, self.model_name = app_name.split('.')
        else:
            self.app_name = app_name
            self.model_name = model_name
        self.content_type_field = content_type_field
        self.id_field = id_field

        super().__init__()

    def get_related_objects(self, obj):
        return self.model.objects.filter(
            **{self.content_type_field: ContentType.objects.get_for_model(obj), self.id_field: obj.pk}
        )

    @property
    def model(self):
        return apps.get_model(self.app_name, self.model_name)


class GenericRelationAnonymizer(RelationAnonymizer):
    """Defines relation for anonymizer to cope with GenericForeignKey."""

    app_name: str
    model_name: str
    content_object_field: str

    def __init__(self, app_name: str, model_name: Optional[str] = None, content_object_field: str = 'content_object'):
        """

        :param app_name: The name of the app or `<app_name>.<model_name>`
        :param model_name: The name of the model with GenericRelation
        :param content_type_field: The name of the FK to ContentType Model
        :param id_field: The id of the related model
        """
        if model_name is None:
            self.app_name, self.model_name = app_name.split('.')
        else:
            self.app_name = app_name
            self.model_name = model_name
        self.content_object_field = content_object_field

        super().__init__()

    def get_related_objects(self, obj):
        model: Model = self.model
        content_obj = getattr(obj, self.content_object_field, None)
        if content_obj is None:
            return model.objects.none()
        if isinstance(content_obj, model):
            return [content_obj]
        return model.objects.none()

    @property
    def model(self):
        return apps.get_model(self.app_name, self.model_name)



from typing import Callable

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model
from django.test import TestCase

from gdpr.models import AnonymizedData
from germanium.tools import assert_false, assert_true


class NotImplementedMixin(TestCase):
    def assertNotImplemented(self, func: Callable, *args, **kwargs) -> None:
        try:
            func(*args, **kwargs)
        except AssertionError as exc:
            print("NOT IMPLEMENTED:", self.id(), exc)
        else:
            raise AssertionError("Function Implemented successfully!!")

    def assertNotImplementedNotEqual(self, *args, **kwargs):
        self.assertNotImplemented(self.assertNotEqual, *args, **kwargs)


class AnonymizedDataMixin(TestCase):
    def assertAnonymizedDataExists(self, obj: Model, field: str):
        content_type = ContentType.objects.get_for_model(obj.__class__)
        assert_true(
            AnonymizedData.objects.filter(content_type=content_type, object_id=str(obj.pk), field=field).exists())

    def assertAnonymizedDataNotExists(self, obj: Model, field: str):
        content_type = ContentType.objects.get_for_model(obj.__class__)
        assert_false(
            AnonymizedData.objects.filter(content_type=content_type, object_id=str(obj.pk), field=field).exists())



from dateutil.relativedelta import relativedelta

from gdpr.purposes.default import AbstractPurpose

from .models import Customer


# SLUG can be any length up to 100 characters
FIRST_AND_LAST_NAME_SLUG = "FNL"
EMAIL_SLUG = "EML"
PAYMENT_VALUE_SLUG = "PVL"
ACCOUNT_SLUG = "ACC"
ACCOUNT_AND_PAYMENT_SLUG = "ACP"
ADDRESS_SLUG = "ADD"
CONTACT_FORM_SLUG = "CTF"
EVERYTHING_SLUG = "EVR"
MARKETING_SLUG = "MKT"
FACEBOOK_SLUG = "FACEBOOK"


class FirstNLastNamePurpose(AbstractPurpose):
    """Store First & Last name for 10 years."""
    name = "retain due to internet archive"
    slug = FIRST_AND_LAST_NAME_SLUG
    expiration_timedelta = relativedelta(years=10)
    fields = ("first_name", "last_name")


class EmailsPurpose(AbstractPurpose):
    """Store emails for 5 years."""
    name = "retain due to over cat overlords"
    slug = EMAIL_SLUG
    expiration_timedelta = relativedelta(years=5)
    fields = (
        ("emails", (
            "email",
        )),
        ("other_registrations", (
            'email_address',
        )),
        ("last_registration", (
            'email_address',
        )),
    )


class PaymentValuePurpose(AbstractPurpose):
    name = "retain due to Foo bar"
    slug = PAYMENT_VALUE_SLUG
    expiration_timedelta = relativedelta(months=6)
    fields = (
        ("accounts", (
            ("payments", (
                "value",
            )),
        )),
    )


class AccountPurpose(AbstractPurpose):
    name = "retain due to Lorem ipsum"
    slug = ACCOUNT_SLUG
    expiration_timedelta = relativedelta(years=2)
    fields = (
        ("accounts", (
            "number",
            "owner"
        )),
    )


class AccountsAndPaymentsPurpose(AbstractPurpose):
    name = "retain due to Gandalf"
    slug = ACCOUNT_AND_PAYMENT_SLUG
    expiration_timedelta = relativedelta(years=3)
    fields = (
        ("accounts", (
            "number",
            "owner",
            ("payments", (
                "value",
                "date"
            )),
        )),
    )


class AddressPurpose(AbstractPurpose):
    name = "retain due to why not?"
    slug = ADDRESS_SLUG
    expiration_timedelta = relativedelta(years=1)
    fields = (
        ("addresses", (
            "street",
            "house_number"
        )),
    )


class EverythingPurpose(AbstractPurpose):
    name = "retain due to Area 51"
    slug = EVERYTHING_SLUG
    expiration_timedelta = relativedelta(years=51)
    fields = (
        "__ALL__",
        ("addresses", "__ALL__"),
        ("accounts", (
            "__ALL__",
            ("payments", (
                "__ALL__",
            ))
        )),
        ("emails", (
            "__ALL__",
        )),
    )


class ContactFormPurpose(AbstractPurpose):
    name = "retain due to mailing campaign"
    slug = CONTACT_FORM_SLUG
    expiration_timedelta = relativedelta(months=1)
    fields = "__ALL__"


class MarketingPurpose(AbstractPurpose):
    """retain due customers wanting more adds"""
    name = "retain due customers wanting more adds"
    slug = MARKETING_SLUG
    expiration_timedelta = relativedelta(years=1)
    fields = ()
    source_model_class = Customer


class FacebookPurpose(AbstractPurpose):
    name = "retain due to facebook ID"
    slug = FACEBOOK_SLUG
    expiration_timedelta = relativedelta(years=5)
    fields = "__ALL__"

    source_model_class = 'tests.Customer'

    def can_anonymize_obj(self, obj, fields):
        return obj.facebook_id is not None



import hashlib
from typing import Any

from gdpr.anonymizers.base import FieldAnonymizer


class BaseHashTextFieldAnonymizer(FieldAnonymizer):
    algorithm: str
    is_reversible = False

    def get_encrypted_value(self, value: Any, encryption_key: str):
        h = hashlib.new(self.algorithm)
        h.update(value.encode('utf-8'))
        return h.hexdigest()[:len(value)] if value else value


class MD5TextFieldAnonymizer(BaseHashTextFieldAnonymizer):
    algorithm = 'md5'


class SHA256TextFieldAnonymizer(BaseHashTextFieldAnonymizer):
    algorithm = 'sha256'


class HashTextFieldAnonymizer(BaseHashTextFieldAnonymizer):

    def __init__(self, algorithm: str, *args, **kwargs):
        if algorithm not in hashlib.algorithms_guaranteed:
            raise RuntimeError(f'Hash algorithm {algorithm} is not supported by python hashlib.')
        self.algorithm = algorithm

        super().__init__(*args, **kwargs)



from datetime import datetime
from dateutil.relativedelta import relativedelta

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from chamber.models import SmartModel

from typing import TYPE_CHECKING, Iterable, Optional, Type

from enumfields import NumEnumField

from .enums import LegalReasonState
from .loading import purpose_register


if TYPE_CHECKING:
    from gdpr.purposes.default import AbstractPurpose


class LegalReasonManager(models.Manager):

    def create_consent(self, purpose_slug: str, source_object, issued_at: Optional[datetime] = None,
                       tag: Optional[str] = None, related_objects: Optional[Iterable[Type[models.Model]]] = None):
        """
        Create (or update, if it exist) a LegalReason with purpose slug for concrete object instance

        Args:
            purpose_slug: String of Legal Reason purpose
            source_object: Source object this Legal Reason is related to
            issued_at: When the Legal Reason consent was given
            tag: String that the developer can add to the created consent and use it to mark his business processes
            related_objects: Objects this Legal Reason relates to (ie. order, registrations etc.)

        Returns:
            Legal Reason: LegalReason object
        """
        try:
            purpose = purpose_register[purpose_slug]
        except KeyError:
            raise KeyError('Purpose with slug {} does not exits'.format(purpose_slug))

        if purpose.source_model_class is not None and not source_object.__class__ == purpose.source_model_class:
            raise KeyError(
                'Purpose with slug {} cannot be created for model {}'.format(purpose_slug, source_object.__class__)
            )

        issued_at = issued_at or timezone.now()

        legal_reason, created = LegalReason.objects.get_or_create(
            source_object_content_type=ContentType.objects.get_for_model(source_object.__class__),
            source_object_id=str(source_object.pk),
            purpose_slug=purpose_slug,
            defaults={
                'issued_at': issued_at,
                'expires_at': issued_at + purpose.expiration_timedelta,
                'tag': tag,
                'state': LegalReasonState.ACTIVE,
            }
        )

        if not created:
            legal_reason.change_and_save(
                expires_at=timezone.now() + purpose.expiration_timedelta,
                tag=tag,
                state=LegalReasonState.ACTIVE
            )

        for related_object in related_objects or ():
            legal_reason.related_objects.update_or_create(
                object_content_type=ContentType.objects.get_for_model(related_object.__class__),
                object_id=related_object.pk
            )

        return legal_reason

    def deactivate_consent(self, purpose_slug: str, source_object):
        """
        Deactivate/Remove consent (Legal reason) for source_object, purpose_slug combination

        Args:
            purpose_slug: Purpose slug to deactivate consent for
            source_object: Source object to deactivate consent for

        Returns:
            List of LegalReason objects
        """
        reasons = []
        for reason in LegalReason.objects.filter_source_instance_active_non_expired_purpose(source_object,
                                                                                            purpose_slug):
            reason.deactivate()
            reasons.append(reason)
        return reasons

    def exists_valid_consent(self, purpose_slug: str, source_object):
        """
        Returns True if source_object has valid (ie. active and non-expired) consent (Legal Reason)

        Args:
            purpose_slug: Purpose_slug to check consent for
            source_object: Source object to check consent for
        """
        return LegalReason.objects.filter_source_instance_active_non_expired_purpose(
            source_object, purpose_slug).exists()

    def exists_deactivated_consent(self, purpose_slug: str, source_object):
        """
        Returns True if source_object has deactivated consent (Legal Reason)

        Args:
            purpose_slug: Purpose_slug to check consent for
            source_object: Source object to check consent for
        """
        return self.filter_source_instance(source_object).filter(
            state=LegalReasonState.DEACTIVATED,
            purpose_slug=purpose_slug
        ).exists()

    def expire_old_consents(self):
        """
        Anonymize and expire consents which have past their `expires_at`.
        """
        for reason in LegalReason.objects.filter_active_and_expired():
            reason.expire()


class LegalReasonQuerySet(models.QuerySet):

    def filter_expired_retaining_data_in_last_days(self, days=None):
        """
        Filters all Legal Reason that retain data and that expired in last days

        Args:
            days: Number of days in the past. If not provided, all Legal Reasons retaining data which expired in the
            past will be returned.
        """
        purpose_slugs_retaining_data = [slug for slug, cls in purpose_register.items() if cls.fields]

        filter_keys = {
            'expires_at__lt': timezone.now(),
        } if days is None else {
            'expires_at__gt': timezone.now() - relativedelta(days=days),
            'expires_at__lt': timezone.now()
        }

        return self.filter(state=LegalReasonState.ACTIVE, purpose_slug__in=purpose_slugs_retaining_data, **filter_keys)

    def filter_non_expired(self):
        return self.filter(Q(expires_at__gte=timezone.now()) | Q(expires_at=None))

    def filter_expired(self):
        return self.filter(expires_at__lte=timezone.now())

    def filter_active(self):
        return self.filter(state=LegalReasonState.ACTIVE)

    def filter_active_and_non_expired(self):
        return self.filter_active().filter_non_expired()

    def filter_active_and_expired(self):
        return self.filter_active().filter_expired()

    def filter_source_instance(self, source_object):
        return self.filter(
            source_object_content_type=ContentType.objects.get_for_model(source_object.__class__),
            source_object_id=str(source_object.pk)
        )

    def filter_source_instance_active_non_expired(self, source_object):
        return self.filter_source_instance(source_object).filter_active_and_non_expired()

    def filter_source_instance_active_non_expired_purpose(self, source_object, purpose_slug: str):
        return self.filter_source_instance_active_non_expired(source_object).filter(
            purpose_slug=purpose_slug
        )


class LegalReason(SmartModel):
    objects = LegalReasonManager.from_queryset(LegalReasonQuerySet)()

    issued_at = models.DateTimeField(
        verbose_name=_('issued at'),
        null=False,
        blank=False,
    )
    expires_at = models.DateTimeField(
        verbose_name=_('expires at'),
        null=True,
        blank=True,
        db_index=True
    )
    tag = models.CharField(
        verbose_name=_('tag'),
        null=True,
        blank=True,
        max_length=100
    )
    state = NumEnumField(
        verbose_name=_('state'),
        null=False,
        blank=False,
        enum=LegalReasonState,
        default=LegalReasonState.ACTIVE,
        db_index=True
    )
    purpose_slug = models.CharField(
        verbose_name=_('purpose'),
        null=False,
        blank=False,
        max_length=100,
        db_index=True
    )
    source_object_content_type = models.ForeignKey(
        ContentType,
        verbose_name=_('source object content type'),
        null=False,
        blank=False,
        on_delete=models.DO_NOTHING
    )
    source_object_id = models.TextField(
        verbose_name=_('source object ID'),
        null=False, blank=False,
        db_index=True
    )
    source_object = GenericForeignKey(
        'source_object_content_type', 'source_object_id'
    )

    class Meta:
        verbose_name = _('legal reason')
        verbose_name_plural = _('legal reasons')
        ordering = ('-created_at',)
        unique_together = ('purpose_slug', 'source_object_content_type', 'source_object_id')

    def __str__(self):
        return f'{self.purpose.name}'

    @property
    def is_active(self):
        return self.state == LegalReasonState.ACTIVE

    @property
    def purpose(self) -> Type["AbstractPurpose"]:
        return purpose_register.get(self.purpose_slug, None)

    def _anonymize_obj(self, *args, **kwargs):
        purpose_register[self.purpose_slug]().anonymize_obj(self.source_object, self, *args, **kwargs)

    def _deanonymize_obj(self, *args, **kwargs):
        purpose_register[self.purpose_slug]().deanonymize_obj(self.source_object, *args, **kwargs)

    def expire(self):
        """Anonymize obj and set state as expired."""
        with transaction.atomic():
            self._anonymize_obj()
            self.change_and_save(state=LegalReasonState.EXPIRED)

    def deactivate(self):
        """Deactivate obj and run anonymization."""
        with transaction.atomic():
            self._anonymize_obj()
            self.change_and_save(state=LegalReasonState.DEACTIVATED)

    def renew(self):
        with transaction.atomic():
            self.change_and_save(
                expires_at=timezone.now() + purpose_register[self.purpose_slug]().expiration_timedelta,
                state=LegalReasonState.ACTIVE
            )
            self._deanonymize_obj()


class LegalReasonRelatedObject(SmartModel):
    legal_reason = models.ForeignKey(
        LegalReason,
        verbose_name=_('legal reason'),
        null=False,
        blank=False,
        related_name='related_objects',
        on_delete=models.CASCADE
    )
    object_content_type = models.ForeignKey(
        ContentType,
        verbose_name=_('related object content type'),
        null=False,
        blank=False,
        on_delete=models.DO_NOTHING
    )
    object_id = models.TextField(
        verbose_name=_('related object ID'),
        null=False,
        blank=False,
        db_index=True
    )
    object = GenericForeignKey(
        'object_content_type', 'object_id'
    )

    class Meta:
        verbose_name = _('legal reason related object')
        verbose_name_plural = _('legal reasons related objects')
        ordering = ('-created_at',)
        unique_together = ('legal_reason', 'object_content_type', 'object_id')

    def __str__(self):
        return '{legal_reason} {object}'.format(legal_reason=self.legal_reason, object=self.object)


class AnonymizedDataQuerySet(models.QuerySet):

    def filter_source_instance_active(self, source_object):
        return self.filter(
            content_type=ContentType.objects.get_for_model(source_object.__class__),
            object_id=str(source_object.pk),
            is_active=True
        )


class AnonymizedData(SmartModel):
    objects = models.Manager.from_queryset(AnonymizedDataQuerySet)()

    field = models.CharField(
        verbose_name=_('anonymized field name'),
        max_length=250,
        null=False,
        blank=False
    )
    content_type = models.ForeignKey(
        ContentType,
        verbose_name=_('related object content type'),
        null=False,
        blank=False,
        on_delete=models.DO_NOTHING
    )
    object_id = models.TextField(
        verbose_name=_('related object ID'),
        null=False,
        blank=False
    )
    object = GenericForeignKey(
        'content_type', 'object_id'
    )
    is_active = models.BooleanField(
        verbose_name=_('is active'),
        default=True
    )
    expired_reason = models.ForeignKey(
        LegalReason,
        verbose_name=_('expired reason'),
        null=True,
        blank=True,
        on_delete=models.SET_NULL
    )

    class Meta:
        verbose_name = _('anonymized data')
        verbose_name_plural = _('anonymized data')
        ordering = ('-created_at',)
        unique_together = ('content_type', 'object_id', 'field')

    def __str__(self):
        return '{field} {object}'.format(field=self.field, object=self.object)



from datetime import timedelta
from typing import List
from unittest import skipIf

from germanium.tools import assert_dict_equal, assert_equal, assert_not_equal, assert_raises

from django.contrib.auth.models import User
from django.core.files.base import ContentFile
from django.test import TestCase

from gdpr.anonymizers import ModelAnonymizer
from gdpr.loading import anonymizer_register
from gdpr.models import LegalReason
from gdpr.utils import (
    get_all_obj_and_parent_versions, get_all_obj_and_parent_versions_queryset_list, get_all_parent_objects,
    get_reversion_local_field_dict, get_reversion_versions, is_reversion_installed
)
from tests.anonymizers import ChildEAnonymizer, ContactFormAnonymizer
from tests.models import (
    Account, Address, Avatar, ChildE, ContactForm, Customer, CustomerRegistration, Email, ExtraParentD, Note, ParentB,
    ParentC, Payment, TopParentA
)
from tests.purposes import EMAIL_SLUG

from .data import (
    ACCOUNT__IBAN, ACCOUNT__NUMBER, ACCOUNT__OWNER, ACCOUNT__SWIFT, ADDRESS__CITY, ADDRESS__HOUSE_NUMBER,
    ADDRESS__POST_CODE, ADDRESS__STREET, CUSTOMER__BIRTH_DATE, CUSTOMER__EMAIL, CUSTOMER__EMAIL2, CUSTOMER__EMAIL3,
    CUSTOMER__FACEBOOK_ID, CUSTOMER__FIRST_NAME, CUSTOMER__IP, CUSTOMER__KWARGS, CUSTOMER__LAST_NAME,
    CUSTOMER__PERSONAL_ID, CUSTOMER__PHONE_NUMBER, PAYMENT__VALUE
)
from .utils import AnonymizedDataMixin, NotImplementedMixin


class TestModelAnonymization(AnonymizedDataMixin, NotImplementedMixin, TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.customer: Customer = Customer(**CUSTOMER__KWARGS)
        cls.customer.save()
        cls.base_encryption_key = 'LoremIpsum'

    def test_anonymize_customer(self):
        self.customer._anonymize_obj()
        anon_customer: Customer = Customer.objects.get(pk=self.customer.pk)

        assert_not_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataExists(anon_customer, 'first_name')
        assert_not_equal(anon_customer.last_name, CUSTOMER__LAST_NAME)
        self.assertAnonymizedDataExists(anon_customer, 'last_name')
        assert_not_equal(anon_customer.full_name, '%s %s' % (CUSTOMER__FIRST_NAME, CUSTOMER__LAST_NAME))
        self.assertAnonymizedDataExists(anon_customer, 'full_name')
        assert_not_equal(anon_customer.primary_email_address, CUSTOMER__EMAIL)
        self.assertAnonymizedDataExists(anon_customer, 'primary_email_address')
        assert_not_equal(anon_customer.personal_id, CUSTOMER__PERSONAL_ID)
        self.assertAnonymizedDataExists(anon_customer, 'personal_id')
        assert_not_equal(anon_customer.phone_number, CUSTOMER__PHONE_NUMBER)
        self.assertAnonymizedDataExists(anon_customer, 'phone_number')
        assert_not_equal(anon_customer.birth_date, CUSTOMER__BIRTH_DATE)
        self.assertAnonymizedDataExists(anon_customer, 'first_name')
        assert_not_equal(anon_customer.facebook_id, CUSTOMER__FACEBOOK_ID)
        self.assertAnonymizedDataExists(anon_customer, 'first_name')
        assert_not_equal(str(anon_customer.last_login_ip), CUSTOMER__IP)
        self.assertAnonymizedDataExists(anon_customer, 'first_name')

    def test_anonymize_customer_registrations(self):
        other_registration: CustomerRegistration = CustomerRegistration.objects.create(email_address=CUSTOMER__EMAIL)
        last_registration: CustomerRegistration = CustomerRegistration.objects.create(email_address=CUSTOMER__EMAIL)
        legal_reason = LegalReason.objects.create_consent(EMAIL_SLUG, self.customer)
        legal_reason.expire()
        other_registration.refresh_from_db()
        last_registration.refresh_from_db()
        assert_not_equal(other_registration.email_address, CUSTOMER__EMAIL)
        assert_not_equal(last_registration.email_address, CUSTOMER__EMAIL)

    def test_email(self):
        self.email: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL)
        self.email.save()
        self.email._anonymize_obj(base_encryption_key=self.base_encryption_key)
        anon_email: Email = Email.objects.get(pk=self.email.pk)

        assert_not_equal(anon_email.email, CUSTOMER__EMAIL)

    def test_address(self):
        self.address: Address = Address(
            customer=self.customer,
            street=ADDRESS__STREET,
            house_number=ADDRESS__HOUSE_NUMBER,
            city=ADDRESS__CITY,
            post_code=ADDRESS__POST_CODE,
        )
        self.address.save()
        self.address._anonymize_obj(base_encryption_key=self.base_encryption_key)
        anon_address: Address = Address.objects.get(pk=self.address.pk)

        assert_not_equal(anon_address.street, ADDRESS__STREET)
        self.assertAnonymizedDataExists(anon_address, 'street')
        assert_equal(anon_address.house_number, ADDRESS__HOUSE_NUMBER)
        assert_equal(anon_address.city, ADDRESS__CITY)
        assert_equal(anon_address.post_code, ADDRESS__POST_CODE)

    def test_account(self):
        self.account: Account = Account(
            customer=self.customer,
            number=ACCOUNT__NUMBER,
            owner=ACCOUNT__OWNER,
            IBAN=ACCOUNT__IBAN,
            swift=ACCOUNT__SWIFT,
        )
        self.account.save()
        self.account._anonymize_obj(base_encryption_key=self.base_encryption_key)

        anon_account: Account = Account.objects.get(pk=self.account.pk)

        assert_not_equal(anon_account.number, ACCOUNT__NUMBER)
        self.assertAnonymizedDataExists(anon_account, 'number')
        assert_not_equal(anon_account.owner, ACCOUNT__OWNER)
        self.assertAnonymizedDataExists(anon_account, 'owner')
        assert_not_equal(anon_account.IBAN, ACCOUNT__IBAN)
        self.assertAnonymizedDataExists(anon_account, 'IBAN')

    def test_payment(self):
        self.account: Account = Account(
            customer=self.customer,
            number=ACCOUNT__NUMBER,
            owner=ACCOUNT__OWNER,
            IBAN=ACCOUNT__IBAN,
            swift=ACCOUNT__SWIFT,
        )
        self.account.save()
        self.payment: Payment = Payment(
            account=self.account,
            value=PAYMENT__VALUE,
        )
        self.payment.save()
        payment_date = self.payment.date

        self.payment._anonymize_obj(base_encryption_key=self.base_encryption_key)

        anon_payment: Payment = Payment.objects.get(pk=self.payment.pk)

        assert_not_equal(anon_payment.value, PAYMENT__VALUE)
        self.assertAnonymizedDataExists(anon_payment, 'value')
        assert_not_equal(anon_payment.date, payment_date)
        self.assertAnonymizedDataExists(anon_payment, 'date')

    def test_contact_form(self):
        FULL_NAME = '%s %s' % (CUSTOMER__FIRST_NAME, CUSTOMER__LAST_NAME)
        self.contact_form: ContactForm = ContactForm(
            email=CUSTOMER__EMAIL,
            full_name=FULL_NAME
        )
        self.contact_form.save()
        self.contact_form._anonymize_obj()

        anon_contact_form: ContactForm = ContactForm.objects.get(pk=self.contact_form.pk)

        assert_not_equal(anon_contact_form.email, CUSTOMER__EMAIL)
        self.assertAnonymizedDataExists(anon_contact_form, 'email')
        assert_not_equal(anon_contact_form.full_name, FULL_NAME)
        self.assertAnonymizedDataExists(anon_contact_form, 'full_name')

    def test_anonymization_of_anonymized_data(self):
        '''Test that anonymized data are not anonymized again.'''
        self.customer._anonymize_obj()
        anon_customer: Customer = Customer.objects.get(pk=self.customer.pk)

        assert_not_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataExists(anon_customer, 'first_name')

        anon_customer._anonymize_obj()
        anon_customer2: Customer = Customer.objects.get(pk=self.customer.pk)

        assert_equal(anon_customer2.first_name, anon_customer.first_name)
        assert_not_equal(anon_customer2.first_name, CUSTOMER__FIRST_NAME)

    def test_anonymization_field_matrix(self):
        self.customer._anonymize_obj(fields=('first_name',))
        anon_customer: Customer = Customer.objects.get(pk=self.customer.pk)

        assert_not_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataExists(anon_customer, 'first_name')

        assert_equal(anon_customer.last_name, CUSTOMER__LAST_NAME)
        self.assertAnonymizedDataNotExists(anon_customer, 'last_name')

    def test_anonymization_field_matrix_related(self):
        related_email: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL)
        related_email.save()

        self.customer._anonymize_obj(fields=('first_name', ('emails', ('email',))))
        anon_customer: Customer = Customer.objects.get(pk=self.customer.pk)

        assert_not_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataExists(anon_customer, 'first_name')

        assert_equal(anon_customer.last_name, CUSTOMER__LAST_NAME)
        self.assertAnonymizedDataNotExists(anon_customer, 'last_name')

        anon_related_email: Email = Email.objects.get(pk=related_email.pk)

        assert_not_equal(anon_related_email.email, CUSTOMER__EMAIL)
        self.assertAnonymizedDataExists(anon_related_email, 'email')

    def test_anonymization_field_matrix_related_all(self):
        related_email: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL)
        related_email.save()

        self.customer._anonymize_obj(fields=('first_name', ('emails', '__ALL__')))
        anon_customer: Customer = Customer.objects.get(pk=self.customer.pk)

        assert_not_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataExists(anon_customer, 'first_name')

        assert_equal(anon_customer.last_name, CUSTOMER__LAST_NAME)
        self.assertAnonymizedDataNotExists(anon_customer, 'last_name')

        anon_related_email: Email = Email.objects.get(pk=related_email.pk)

        assert_not_equal(anon_related_email.email, CUSTOMER__EMAIL)
        self.assertAnonymizedDataExists(anon_related_email, 'email')

    def test_reverse_generic_relation(self):
        note: Note = Note(note='Test message')
        note.content_object = self.customer
        note.save()

        self.customer._anonymize_obj(fields=(('notes', '__ALL__'),))

        anon_note: Note = Note.objects.get(pk=note.pk)

        assert_not_equal(anon_note.note, note.note)
        self.assertAnonymizedDataExists(note, 'note')

        self.customer._deanonymize_obj(fields=(('notes', '__ALL__'),))

        anon_note2: Note = Note.objects.get(pk=note.pk)

        assert_equal(anon_note2.note, note.note)
        self.assertAnonymizedDataNotExists(note, 'note')

    def test_irreversible_deanonymization(self):
        contact_form: ContactForm = ContactForm(email=CUSTOMER__EMAIL, full_name=CUSTOMER__LAST_NAME)
        contact_form.save()
        contact_form._anonymize_obj(fields=('__ALL__',))

        assert_raises(ModelAnonymizer.IrreversibleAnonymizerException, contact_form._deanonymize_obj,
                      fields=('__ALL__',))

    def test_generic_relation_anonymizer(self):
        contact_form: ContactForm = ContactForm(email=CUSTOMER__EMAIL, full_name=CUSTOMER__LAST_NAME)
        contact_form.save()
        note: Note = Note(note='Test message')
        note.content_object = contact_form
        note.save()

        note._anonymize_obj(fields=(('contact_form', '__ALL__'),), base_encryption_key=self.base_encryption_key)

        anon_contact_form: ContactForm = ContactForm.objects.get(pk=contact_form.pk)

        assert_not_equal(anon_contact_form.email, CUSTOMER__EMAIL)
        self.assertAnonymizedDataExists(anon_contact_form, 'email')
        assert_not_equal(anon_contact_form.full_name, CUSTOMER__LAST_NAME)
        self.assertAnonymizedDataExists(anon_contact_form, 'full_name')

    @skipIf(not is_reversion_installed(), 'Django-reversion is not installed.')
    def test_reversion_anonymization(self):
        from reversion import revisions as reversion
        from reversion.models import Version
        from gdpr.utils import get_reversion_versions

        anon = ContactFormAnonymizer()
        anon.Meta.anonymize_reversion = True
        anon.Meta.reversible_anonymization = True

        user = User(username='test_username')
        user.save()

        with reversion.create_revision():
            form = ContactForm()
            form.email = CUSTOMER__EMAIL
            form.full_name = CUSTOMER__LAST_NAME
            form.save()

            reversion.set_user(user)

        with reversion.create_revision():
            form.email = CUSTOMER__EMAIL2
            form.save()

            reversion.set_user(user)

        with reversion.create_revision():
            form.email = CUSTOMER__EMAIL3
            form.save()

            reversion.set_user(user)

        versions: List[Version] = get_reversion_versions(form).order_by('id')

        assert_equal(versions[0].field_dict['email'], CUSTOMER__EMAIL)
        assert_equal(versions[1].field_dict['email'], CUSTOMER__EMAIL2)
        assert_equal(versions[2].field_dict['email'], CUSTOMER__EMAIL3)

        anon.anonymize_obj(form, base_encryption_key=self.base_encryption_key)

        anon_versions: List[Version] = get_reversion_versions(form).order_by('id')
        anon_form = ContactForm.objects.get(pk=form.pk)

        assert_not_equal(anon_versions[0].field_dict['email'], CUSTOMER__EMAIL)
        assert_not_equal(anon_versions[1].field_dict['email'], CUSTOMER__EMAIL2)
        assert_not_equal(anon_versions[2].field_dict['email'], CUSTOMER__EMAIL3)
        assert_not_equal(anon_form.email, CUSTOMER__EMAIL3)

        anon.deanonymize_obj(anon_form, base_encryption_key=self.base_encryption_key)

        deanon_versions: List[Version] = get_reversion_versions(form).order_by('id')
        deanon_form = ContactForm.objects.get(pk=form.pk)

        assert_equal(deanon_versions[0].field_dict['email'], CUSTOMER__EMAIL)
        assert_equal(deanon_versions[1].field_dict['email'], CUSTOMER__EMAIL2)
        assert_equal(deanon_versions[2].field_dict['email'], CUSTOMER__EMAIL3)
        assert_equal(deanon_form.email, CUSTOMER__EMAIL3)
        assert_dict_equal(versions[0].field_dict, deanon_versions[0].field_dict)
        assert_dict_equal(versions[1].field_dict, deanon_versions[1].field_dict)
        assert_dict_equal(versions[2].field_dict, deanon_versions[2].field_dict)

    @skipIf(not is_reversion_installed(), 'Django-reversion is not installed.')
    def test_reversion_delete(self):
        from reversion import revisions as reversion
        from reversion.models import Version
        from gdpr.utils import get_reversion_versions

        anon = ContactFormAnonymizer()
        anon.Meta.delete_reversion = True

        user = User(username='test_username')
        user.save()

        with reversion.create_revision():
            form = ContactForm()
            form.email = CUSTOMER__EMAIL
            form.full_name = CUSTOMER__LAST_NAME
            form.save()

            reversion.set_user(user)

        with reversion.create_revision():
            form.email = CUSTOMER__EMAIL2
            form.save()

            reversion.set_user(user)

        with reversion.create_revision():
            form.email = CUSTOMER__EMAIL3
            form.save()

            reversion.set_user(user)

        versions: List[Version] = get_reversion_versions(form).order_by('id')

        assert_equal(versions[0].field_dict['email'], CUSTOMER__EMAIL)
        assert_equal(versions[1].field_dict['email'], CUSTOMER__EMAIL2)
        assert_equal(versions[2].field_dict['email'], CUSTOMER__EMAIL3)

        anon.anonymize_obj(form, base_encryption_key=self.base_encryption_key)

        anon_versions: List[Version] = get_reversion_versions(form).order_by('id')
        assert_equal(len(anon_versions), 0)

    def test_get_all_parent_objects(self):
        e = ChildE.objects.create(
            name='Lorem',
            first_name='Ipsum',
            last_name='Dolor',
            birth_date=CUSTOMER__BIRTH_DATE,
            note='sit Amet'
        )

        parent_objects = get_all_parent_objects(e)
        assert_equal(len(parent_objects), 4)
        assert_equal([obj.__class__ for obj in parent_objects], [ParentC, ExtraParentD, ParentB, TopParentA])

    @skipIf(not is_reversion_installed(), 'Django-reversion is not installed.')
    def test_get_all_obj_and_parent_versions_queryset_list(self):
        from reversion import revisions as reversion

        with reversion.create_revision():
            e = ChildE()
            e.name = 'Lorem'
            e.first_name = 'Ipsum'
            e.last_name = 'Dolor'
            e.birth_date = CUSTOMER__BIRTH_DATE
            e.note = 'sit Amet'
            e.save()

        versions_queryset_list = get_all_obj_and_parent_versions_queryset_list(e)
        assert_equal(len(versions_queryset_list), 5)
        assert_equal(
            [qs[0]._content_type.model_class() for qs in versions_queryset_list],
            [ParentC, ExtraParentD, ParentB, TopParentA, ChildE]
        )

    @skipIf(not is_reversion_installed(), 'Django-reversion is not installed.')
    def test_get_all_obj_and_parent_versions(self):
        from reversion import revisions as reversion

        with reversion.create_revision():
            e = ChildE()
            e.name = 'Lorem'
            e.first_name = 'Ipsum'
            e.last_name = 'Dolor'
            e.birth_date = CUSTOMER__BIRTH_DATE
            e.note = 'sit Amet'
            e.save()

        versions_list = get_all_obj_and_parent_versions(e)
        assert_equal(len(versions_list), 5)
        assert_equal(
            [version._content_type.model_class() for version in versions_list],
            [ParentC, ExtraParentD, ParentB, TopParentA, ChildE]
        )

    @skipIf(not is_reversion_installed(), 'Django-reversion is not installed.')
    def test_reversion_anonymization_parents(self):
        from reversion import revisions as reversion
        from reversion.models import Version

        anon = ChildEAnonymizer()

        user = User(username='testing_username')
        user.save()

        with reversion.create_revision():
            e = ChildE()
            e.name = 'Lorem'
            e.first_name = 'Ipsum'
            e.last_name = 'Dolor'
            e.birth_date = CUSTOMER__BIRTH_DATE
            e.note = 'sit Amet'
            e.save()

            reversion.set_user(user)

        with reversion.create_revision():
            e.name = 'LOREM'
            e.first_name = 'IPSUM'
            e.last_name = 'DOLOR'
            e.birth_date = CUSTOMER__BIRTH_DATE + timedelta(days=2)
            e.note = 'SIT AMET'
            e.save()

            reversion.set_user(user)

        versions_a: List[Version] = get_reversion_versions(e.topparenta_ptr).order_by('id')
        versions_b: List[Version] = get_reversion_versions(e.parentb_ptr).order_by('id')
        versions_c: List[Version] = get_reversion_versions(e.parentc_ptr).order_by('id')
        versions_d: List[Version] = get_reversion_versions(e.extraparentd_ptr).order_by('id')
        versions_e: List[Version] = get_reversion_versions(e).order_by('id')

        assert_equal(get_reversion_local_field_dict(versions_a[0])['name'], 'Lorem')
        assert_equal(get_reversion_local_field_dict(versions_a[1])['name'], 'LOREM')

        assert_equal(get_reversion_local_field_dict(versions_b[0])['birth_date'], CUSTOMER__BIRTH_DATE)
        assert_equal(get_reversion_local_field_dict(versions_b[1])['birth_date'],
                     CUSTOMER__BIRTH_DATE + timedelta(days=2))

        assert_equal(get_reversion_local_field_dict(versions_c[0])['first_name'], 'Ipsum')
        assert_equal(get_reversion_local_field_dict(versions_c[1])['first_name'], 'IPSUM')

        assert_equal(get_reversion_local_field_dict(versions_d[0])['note'], 'sit Amet')
        assert_equal(get_reversion_local_field_dict(versions_d[1])['note'], 'SIT AMET')

        assert_equal(get_reversion_local_field_dict(versions_e[0])['last_name'], 'Dolor')
        assert_equal(get_reversion_local_field_dict(versions_e[1])['last_name'], 'DOLOR')

        anon.anonymize_obj(e, base_encryption_key=self.base_encryption_key)

        anon_versions_a: List[Version] = get_reversion_versions(e.topparenta_ptr).order_by('id')
        anon_versions_b: List[Version] = get_reversion_versions(e.parentb_ptr).order_by('id')
        anon_versions_c: List[Version] = get_reversion_versions(e.parentc_ptr).order_by('id')
        anon_versions_d: List[Version] = get_reversion_versions(e.extraparentd_ptr).order_by('id')
        anon_versions_e: List[Version] = get_reversion_versions(e).order_by('id')
        anon_e = ChildE.objects.get(pk=e.pk)

        assert_not_equal(get_reversion_local_field_dict(anon_versions_a[0])['name'], 'Lorem')
        assert_not_equal(get_reversion_local_field_dict(anon_versions_a[1])['name'], 'LOREM')

        assert_not_equal(get_reversion_local_field_dict(anon_versions_b[0])['birth_date'], CUSTOMER__BIRTH_DATE)
        assert_not_equal(get_reversion_local_field_dict(anon_versions_b[1])['birth_date'],
                         CUSTOMER__BIRTH_DATE + timedelta(days=2))

        assert_not_equal(get_reversion_local_field_dict(anon_versions_c[0])['first_name'], 'Ipsum')
        assert_not_equal(get_reversion_local_field_dict(anon_versions_c[1])['first_name'], 'IPSUM')

        assert_not_equal(get_reversion_local_field_dict(anon_versions_d[0])['note'], 'sit Amet')
        assert_not_equal(get_reversion_local_field_dict(anon_versions_d[1])['note'], 'SIT AMET')

        assert_not_equal(get_reversion_local_field_dict(anon_versions_e[0])['last_name'], 'Dolor')
        assert_not_equal(get_reversion_local_field_dict(anon_versions_e[1])['last_name'], 'DOLOR')

        anon.deanonymize_obj(anon_e, base_encryption_key=self.base_encryption_key)

        deanon_versions_a: List[Version] = get_reversion_versions(e.topparenta_ptr).order_by('id')
        deanon_versions_b: List[Version] = get_reversion_versions(e.parentb_ptr).order_by('id')
        deanon_versions_c: List[Version] = get_reversion_versions(e.parentc_ptr).order_by('id')
        deanon_versions_d: List[Version] = get_reversion_versions(e.extraparentd_ptr).order_by('id')
        deanon_versions_e: List[Version] = get_reversion_versions(e).order_by('id')

        assert_equal(get_reversion_local_field_dict(deanon_versions_a[0])['name'], 'Lorem')
        assert_equal(get_reversion_local_field_dict(deanon_versions_a[1])['name'], 'LOREM')

        assert_equal(get_reversion_local_field_dict(deanon_versions_b[0])['birth_date'], CUSTOMER__BIRTH_DATE)
        assert_equal(get_reversion_local_field_dict(deanon_versions_b[1])['birth_date'],
                     CUSTOMER__BIRTH_DATE + timedelta(days=2))

        assert_equal(get_reversion_local_field_dict(deanon_versions_c[0])['first_name'], 'Ipsum')
        assert_equal(get_reversion_local_field_dict(deanon_versions_c[1])['first_name'], 'IPSUM')

        assert_equal(get_reversion_local_field_dict(deanon_versions_d[0])['note'], 'sit Amet')
        assert_equal(get_reversion_local_field_dict(deanon_versions_d[1])['note'], 'SIT AMET')

        assert_equal(get_reversion_local_field_dict(deanon_versions_e[0])['last_name'], 'Dolor')
        assert_equal(get_reversion_local_field_dict(deanon_versions_e[1])['last_name'], 'DOLOR')


class TestFileFieldAnonymizer(TestCase):
    def test_file_field(self):
        customer = Customer(**CUSTOMER__KWARGS)
        customer.save()

        avatar = Avatar()
        avatar.customer = customer
        avatar.image.save('test_file_secret_data', ContentFile('Super secret data'), save=False)
        avatar.save()

        avatar_2: Avatar = Avatar.objects.last()
        assert_equal(avatar_2.image.read(), b'Super secret data')
        avatar_2._anonymize_obj(base_encryption_key='LoremIpsumDolorSitAmet')

        avatar_3: Avatar = Avatar.objects.last()
        assert_not_equal(avatar_3.image.read(), b'Super secret data')

        # Cleanup
        avatar_3.image.delete()
        avatar_3.delete()

    def test_file_field_real_file(self):
        anonymizer = anonymizer_register[Avatar]
        anonymizer.image.replacement_file = 'test_file'
        customer = Customer(**CUSTOMER__KWARGS)
        customer.save()

        avatar = Avatar()
        avatar.customer = customer
        avatar.image.save('test_file_real', ContentFile('Super secret data'))

        avatar_2: Avatar = Avatar.objects.last()
        assert_equal(avatar_2.image.read(), b'Super secret data')
        avatar_2._anonymize_obj(base_encryption_key='LoremIpsumDolorSitAmet')

        avatar_3: Avatar = Avatar.objects.last()
        assert_not_equal(avatar_3.image.read(), b'Super secret data')

        anonymizer.image.replacement_file = None
        # Cleanup
        avatar_3.image.delete()
        avatar_3.delete()



from decimal import Decimal

from django.test import TestCase

from faker import Faker
from gdpr.encryption import (
    decrypt_email_address, decrypt_text, encrypt_email_address, encrypt_text, translate_iban, translate_number
)
from germanium.tools import assert_equal, assert_not_equal


IBANS = [
    'AL47 2121 1009 0000 0002 3569 8741',
    'AD12 0001 2030 2003 5910 0100',
    'AZ21 NABZ 0000 0000 1370 1000 1944',
    'BH67 BMAG 0000 1299 1234 56',
    'BE68 5390 0754 7034',
    'BY13 NBRB 3600 9000 0000 2Z00 AB00',
    'BA39 1290 0794 0102 8494',
    'BR18 0036 0305 0000 1000 9795 493C 1',
    'VG96 VPVG 0000 0123 4567 8901',
    'BG80 BNBG 9661 1020 3456 78',
    'ME25 5050 0001 2345 6789 51',
    'CZ65 0800 0000 1920 0014 5399',
    'DK50 0040 0440 1162 43',
    'DO28 BAGR 0000 0001 2124 5361 1324',
    'EE38 2200 2210 2014 5685',
    'FI21 1234 5600 0007 85',
    'FR14 2004 1010 0505 0001 3M02 606',
    'GI75 NWBK 0000 0000 7099 453',
    'GE29 NB00 0000 0101 9049 17',
    'GT82 TRAJ 0102 0000 0012 1002 9690',
    'HR12 1001 0051 8630 0016 0',
    'IQ98 NBIQ 8501 2345 6789 012',
    'IE29 AIBK 9311 5212 3456 78',
    'IS14 0159 2600 7654 5510 7303 39',
    'IT60 X054 2811 1010 0000 0123 456',
    'IL62 0108 0000 0009 9999 999',
    'JO94 CBJO 0010 0000 0000 0131 0003 02',
    'QA58 DOHB 0000 1234 5678 90AB CDEF G',
    'KZ86 125K ZT50 0410 0100',
    'XK05 1212 0123 4567 8906',
    'CR05 0152 0200 1026 2840 66',
    'KW81 CBKU 0000 0000 0000 1234 5601 01',
    'CY17 0020 0128 0000 0012 0052 7600',
    'LB62 0999 0000 0001 0019 0122 9114',
    'LI21 0881 0000 2324 013A A',
    'LT12 1000 0111 0100 1000',
    'LV80 BANK 0000 4351 9500 1',
    'LU28 0019 4006 4475 0000',
    'HU42 1177 3016 1111 1018 0000 0000',
    'MK07 2501 2000 0058 984',
    'MT84 MALT 0110 0001 2345 MTLC AST0 01S',
    'MU17 BOMM 0101 1010 3030 0200 000M UR',
    'MR13 0002 0001 0100 0012 3456 753',
    'MD24 AG00 0225 1000 1310 4168',
    'MC58 1122 2000 0101 2345 6789 030',
    'DE89 3704 0044 0532 0130 00',
    'NL91 ABNA 0417 1643 00',
    'NO93 8601 1117 947',
    'PK36 SCBL 0000 0011 2345 6702',
    'PS92 PALS 0000 0000 0400 1234 5670 2',
    'PL61 1090 1014 0000 0712 1981 2874',
    'PT50 0002 0123 1234 5678 9015 4',
    'AT61 1904 3002 3457 3201',
    'RO49 AAAA 1B31 0075 9384 0000',
    'GR16 0110 1250 0000 0001 2300 695',
    'SV62 CENR 0000 0000 0000 0070 0025',
    'SM86 U032 2509 8000 0000 0270 100',
    'SA03 8000 0000 6080 1016 7519',
    'SC18 SSCB 1101 0000 0000 0000 1497 USD',
    'SK31 1200 0000 1987 4263 7541',
    'SI56 2633 0001 2039 086',
    'AE07 0331 2345 6789 0123 456',
    'RS35 2600 0560 1001 6113 79',
    'LC55 HEMM 0001 0001 0012 0012 0002 3015',
    'ST68 0001 0001 0051 8453 1011 2',
    'ES91 2100 0418 4502 0005 1332',
    'SE45 5000 0000 0583 9825 7466',
    'CH93 0076 2011 6238 5295 7',
    'TN59 1000 6035 1835 9847 8831',
    'TR33 0006 1005 1978 6457 8413 26',
    'UA21 3223 1300 0002 6007 2335 6600 1',
    'VA59 0011 2300 0012 3456 78',
    'GB29 NWBK 6016 1331 9268 19',
    'TL38 0080 0123 4567 8910 157',
]


class TestEncryption(TestCase):
    """
    Tests the `gdpr.encryption` module.
    """

    def setUp(self):
        self.faker = Faker()
        self.encryption_key = 'LoremIpsumDolorSitAmet'
        self.numeric_encryption_key = '314159265358'

    def test_encrypt_text_full_name(self):
        """
        Test function `gdpr.encryption.encrypt_text` by using human full name from Faker lib.
        """
        cleartext = self.faker.name()

        ciphertext = encrypt_text(self.encryption_key, cleartext)
        assert_not_equal(cleartext, ciphertext, "The encrypted name is equal to the original name.")

        decrypted = decrypt_text(self.encryption_key, ciphertext)
        assert_equal(cleartext, decrypted, "The decrypted name is not equal to the original name.")

    def test_encrypt_email_address(self):
        """
        Test function `gdpr.encryption.encrypt_email_address` by using email address from Faker lib.
        """
        cleartext = self.faker.email()

        ciphertext = encrypt_email_address(self.encryption_key, cleartext)
        assert_not_equal(cleartext, ciphertext, "The encrypted email address is equal to the original email address.")

        decrypted = decrypt_email_address(self.encryption_key, ciphertext)
        assert_equal(cleartext, decrypted, "The decrypted email address is not equal to the original email address.")

    def test_translate_iban(self):
        """
        Test function `gdpr.encryption.translate_iban` by using an example IBAN for every country using IBAN system.
        """
        for IBAN in IBANS:
            encrypted = translate_iban(self.encryption_key, IBAN)
            assert_not_equal(encrypted, IBAN, "The encrypted IBAN is equal to the original IBAN.")
            assert_equal(translate_iban(self.encryption_key, encrypted, False), IBAN,
                         "The decrypted IBAN is not equal to the original IBAN.")

    def test_translate_number_whole_positive(self):
        """
        Test metod `translate_number` on whole positive number.
        """
        number = 42
        encrypted = translate_number(self.numeric_encryption_key, number)

        assert_not_equal(number, encrypted)
        assert_equal(type(number), type(encrypted))

        decrypted = translate_number(self.numeric_encryption_key, encrypted, encrypt=False)

        assert_equal(number, decrypted)
        assert_equal(type(number), type(decrypted))

    def test_translate_number_whole_negative(self):
        """
        Test metod `translate_number` on whole negative number.
        """
        number = -42
        encrypted = translate_number(self.numeric_encryption_key, number)

        assert_not_equal(number, encrypted)
        assert_equal(type(number), type(encrypted))

        decrypted = translate_number(self.numeric_encryption_key, encrypted, encrypt=False)

        assert_equal(number, decrypted)
        assert_equal(type(number), type(decrypted))

    def test_translate_number_decimal_positive(self):
        """
        Test metod `translate_number` on decimal positive number.
        """
        number = Decimal("3.14")
        encrypted = translate_number(self.numeric_encryption_key, number)

        assert_not_equal(number, encrypted)
        assert_equal(type(number), type(encrypted))

        decrypted = translate_number(self.numeric_encryption_key, encrypted, encrypt=False)

        assert_equal(number, decrypted)
        assert_equal(type(number), type(decrypted))

    def test_translate_number_decimal_negative(self):
        """
        Test metod `translate_number` on decimal positive number.
        """
        number = Decimal("-3.14")
        encrypted = translate_number(self.numeric_encryption_key, number)

        assert_not_equal(number, encrypted)
        assert_equal(type(number), type(encrypted))

        decrypted = translate_number(self.numeric_encryption_key, encrypted, encrypt=False)

        assert_equal(number, decrypted)
        assert_equal(type(number), type(decrypted))



# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2019-02-28 13:25
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gdpr', '0005_anon_2_0'),
    ]

    operations = [
        migrations.AlterField(
            model_name='legalreason',
            name='source_object_id',
            field=models.TextField(db_index=True, verbose_name='source object ID'),
        ),
        migrations.AlterField(
            model_name='legalreasonrelatedobject',
            name='object_id',
            field=models.TextField(db_index=True, verbose_name='related object ID'),
        ),
    ]



import re
from datetime import date

from django.core.exceptions import ValidationError
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _


def get_day_from_personal_id(personal_id):
    day = int(personal_id[4:6])
    if day > 50:
        day -= 50
    return day


def get_month_from_personal_id(personal_id):
    year = get_year_from_personal_id(personal_id)
    month = int(personal_id[2:4])
    if month > 70 and year > 2003:
        month -= 70
    elif month > 50:
        month -= 50
    elif month > 20 and year > 2003:
        month -= 20
    return month


def get_year_from_personal_id(personal_id):
    year = int(personal_id[0:2])
    value = personal_id.replace('/', '')
    year += 2000 if year < 54 and len(value) == 10 else 1900
    return year


def personal_id_date(personal_id):
    try:
        return date(get_year_from_personal_id(personal_id), get_month_from_personal_id(personal_id),
                    get_day_from_personal_id(personal_id))
    except ValueError:
        raise ValueError('Invalid personal id')


class CZBirthNumberValidator:
    """
    Czech birth number field validator.
    """
    BIRTH_NUMBER = re.compile(r'^(?P<birth>\d{6})/?(?P<id>\d{3,4})$')

    def __call__(self, value):
        value = force_text(value)

        match = re.match(self.BIRTH_NUMBER, value)
        if not match:
            raise ValidationError(_('Enter a birth number in the format XXXXXX/XXXX.'))

        birth, id = match.groupdict()['birth'], match.groupdict()['id']

        # Three digits for verificatin number were used until 1. january 1954
        if len(id) != 3:
            # Fourth digit has been added since 1. January 1954.
            # It is modulo of dividing birth number and verification number by 11.
            # If the modulo were 10, the last number was 0 (and therefore, the whole
            # birth number weren't dividable by 11. These number are no longer used (since 1985)
            # and condition 'modulo == 10' can be removed in 2085.

            modulo = int(birth + id[:3]) % 11

            if (modulo != int(id[-1])) and (modulo != 10 or id[-1] != '0'):
                raise ValidationError(_('Enter a valid birth number.'))

        try:
            personal_id_date(value)
        except ValueError:
            raise ValidationError(_('Enter a valid birth number.'))


class IDCardNoValidator:
    """
    Czech id card number field validator.
    """
    ID_CARD_NUMBER = re.compile(r'^\d{9}$')

    def __call__(self, value):
        value = force_text(value)

        match = re.match(self.ID_CARD_NUMBER, value)
        if not match:
            raise ValidationError(_('Enter an ID card in the format XXXXXXXXX.'))
        elif value[0] == '0':
            raise ValidationError(_('Enter a valid ID card number.'))
        else:
            return value


class BankAccountValidator:
    BANK_ACCOUNT_NUMBER_REVERSE_PATTERN = re.compile(
        r'^(?P<bank>\d{1,6})/(?P<number>\d{1,10})(-?(?P<prefix>\d{1,6}))?$')

    def __call__(self, value):
        match = re.match(self.BANK_ACCOUNT_NUMBER_REVERSE_PATTERN, force_text(value)[::-1])
        if match:
            return construct_bank_account_number((match.groupdict()['prefix'] or '')[::-1],
                                                 match.groupdict()['number'][::-1],
                                                 match.groupdict()['bank'][::-1])
        else:
            raise ValidationError(_('Enter a valid bank account number.'))


def construct_bank_account_number(prefix, number, bank_code):
    return '{:0>6}-{:0>10}/{}'.format(prefix, number, bank_code)


def split_bank_account_to_prefix_postfix(bank_account_number):
    return bank_account_number.split('-') if '-' in bank_account_number else ('', bank_account_number)


def clean_bank_account_number_or_none(bank_account_number):
    try:
        return BankAccountValidator()(bank_account_number)
    except ValidationError:
        return None



from django.test import TestCase

from gdpr.fields import Fields
from germanium.tools import assert_list_equal, assert_true
from tests.anonymizers import CustomerAnonymizer
from tests.models import Customer


LOCAL_FIELDS = ("first_name", "last_name")

BASIC_FIELDS = (
    "primary_email_address",
    ("emails", (
        "email",
    )),
)

MULTILEVEL_FIELDS = (
    ("accounts", (
        "number",
        "owner",
        ("payments", (
            "value",
            "date"
        ))
    )),
)


class TestFields(TestCase):
    def test_local_all(self):
        fields = Fields('__ALL__', Customer)
        assert_list_equal(fields.local_fields, list(CustomerAnonymizer().fields.keys()))

    def test_local(self):
        fields = Fields(LOCAL_FIELDS, Customer)
        assert_list_equal(fields.local_fields, list(LOCAL_FIELDS))

    def test_local_and_related(self):
        fields = Fields(BASIC_FIELDS, Customer)

        assert_list_equal(fields.local_fields, ['primary_email_address'])
        assert_true('emails' in fields.related_fields)
        assert_list_equal(fields.related_fields['emails'].local_fields, ['email'])

    def test_multilevel_related(self):
        fields = Fields(MULTILEVEL_FIELDS, Customer)

        assert_list_equal(fields.local_fields, [])
        assert_true('accounts' in fields.related_fields)
        assert_list_equal(fields.related_fields['accounts'].local_fields, ['number', 'owner'])
        assert_true('payments' in fields.related_fields['accounts'].related_fields)
        assert_list_equal(fields.related_fields['accounts'].related_fields['payments'].local_fields, ['value', 'date'])



# Generated by Django 3.1 on 2021-01-02 10:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gdpr', '0009_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='legalreason',
            name='purpose_slug',
            field=models.CharField(db_index=True, max_length=100, verbose_name='purpose'),
        ),
    ]



from django.core.exceptions import ValidationError
from django.test import TestCase

from gdpr.anonymizers.local.cs import (
    CzechAccountNumber, CzechAccountNumberFieldAnonymizer, CzechIBAN, CzechIBANSmartFieldAnonymizer,
    CzechIDCardFieldAnonymizer, CzechPersonalIDSmartFieldAnonymizer, CzechPhoneNumberFieldAnonymizer
)
from germanium.tools import assert_equal, assert_false, assert_not_equal, assert_raises, assert_true


class TestCzechAccountNumberField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = CzechAccountNumberFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_account_number_simple_field(self):
        account_number = '2501277007/2010'
        out = self.field.get_encrypted_value(account_number, self.encryption_key)

        assert_not_equal(out, account_number)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, account_number)

    def test_account_number_simple_field_smart_method(self):
        field = CzechAccountNumberFieldAnonymizer(use_smart_method=True)
        account_number = '2501277007/2010'
        out = field.get_encrypted_value(account_number, self.encryption_key)

        assert_not_equal(out, account_number)

        out_decrypt = field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, account_number)

    def test_account_number_with_pre_num_field(self):
        account_number = '19-2000145399/0800'
        out = self.field.get_encrypted_value(account_number, self.encryption_key)

        assert_not_equal(out, account_number)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, account_number)

    def test_account_number_with_pre_num_field_smart_method(self):
        field = CzechAccountNumberFieldAnonymizer(use_smart_method=True)
        account_number = '19-2000145399/0800'
        out = field.get_encrypted_value(account_number, self.encryption_key)

        assert_not_equal(out, account_number)

        out_decrypt = field.get_decrypted_value(out, self.encryption_key)

        assert_equal(out_decrypt, account_number)

    def test_account_format_check(self):
        assert_true(CzechAccountNumber.parse('19-2000145399/0800').check_account_format())
        assert_true(CzechAccountNumber.parse('2501277007/2010').check_account_format())

    def test_brute_force(self):
        account = CzechAccountNumber.parse('19-2000145399/0800')
        key = 314
        original_account_num = account.num

        account.brute_force_next(key)

        assert_not_equal(original_account_num, account.num)

        account.brute_force_prev(key)

        assert_equal(original_account_num, account.num)


class TestCzechIBANSmartFieldAnonymizer(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = CzechIBANSmartFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'
        cls.text_iban = 'CZ65 0800 0000 1920 0014 5399'
        cls.no_space_text_iban = 'CZ6508000000192000145399'
        cls.invalid_text_iban = 'CZ00 0800 0000 1920 0014 5399'
        cls.no_pre_num_iban = 'CZ4601000000000099550247'

    def test_czech_iban_field(self):
        out = self.field.get_encrypted_value(self.text_iban, self.encryption_key)
        assert_not_equal(out, self.text_iban)

        out_decrypted = self.field.get_decrypted_value(out, self.encryption_key)
        assert_equal(out_decrypted, self.text_iban)

    def test_czech_iban_field_no_space(self):
        out = self.field.get_encrypted_value(self.no_space_text_iban, self.encryption_key)
        assert_not_equal(out, self.no_space_text_iban)
        assert_not_equal(out, self.text_iban)

        out_decrypted = self.field.get_decrypted_value(out, self.encryption_key)
        assert_equal(out_decrypted, self.no_space_text_iban)

    def test_czech_iban_field_get_encrypted_value_invalid_format_raises(self):
        assert_raises(ValidationError, self.field.get_encrypted_value, self.invalid_text_iban, self.encryption_key)

    def test_czech_iban_field_get_decrypted_value_invalid_format_raises(self):
        assert_raises(ValidationError, self.field.get_decrypted_value, self.invalid_text_iban, self.encryption_key)

    def test_czech_iban_parse_and_str_with_spaces(self):
        assert_equal(self.text_iban, str(CzechIBAN.parse(self.text_iban)))

    def test_czech_iban_parse_and_str_without_spaces(self):
        assert_equal(self.no_space_text_iban, str(CzechIBAN.parse(self.no_space_text_iban)))

    def test_czech_iban_check_format(self):
        assert_true(CzechIBAN.parse(self.text_iban).check_iban_format())

    def test_czech_iban_check_format_invalid(self):
        assert_false(CzechIBAN.parse(self.invalid_text_iban).check_iban_format())

    def test_czech_iban_check_format_no_pre_num(self):
        assert_true(CzechIBAN.parse(self.no_pre_num_iban).check_iban_format())

    def test_brute_force(self):
        account = CzechIBAN.parse(self.text_iban)
        key = 314

        account.brute_force_next(key)
        assert_not_equal(self.text_iban, str(account))

        account.brute_force_prev(key)
        assert_equal(self.text_iban, str(account))


class TestCzechPhoneNumberField(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = CzechPhoneNumberFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_basic_phone_number(self):
        phone_number = '608104120'
        out = self.field.get_encrypted_value(phone_number, self.encryption_key)

        assert_not_equal(phone_number, out)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(phone_number, out_decrypt)

    def test_plus_area_code_phone_number(self):
        phone_number = '+420608104120'
        out = self.field.get_encrypted_value(phone_number, self.encryption_key)

        assert_not_equal(phone_number, out)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(phone_number, out_decrypt)

    def test_zero_zero_area_code_phone_number(self):
        phone_number = '+420608104120'
        out = self.field.get_encrypted_value(phone_number, self.encryption_key)

        assert_not_equal(phone_number, out)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)

        assert_equal(phone_number, out_decrypt)


class TestCzechIDCardFieldAnonymizer(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = CzechIDCardFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_czech_id_card_field_anonymizer(self):
        id_card = "297065518"

        out = self.field.get_encrypted_value(id_card, self.encryption_key)
        assert_not_equal(id_card, out)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)
        assert_equal(id_card, out_decrypt)


class TestCzechPersonalIDSmartFieldAnonymizer(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = CzechPersonalIDSmartFieldAnonymizer()
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    def test_czech_personal_id_smart_field_anonymizer(self):
        personal_id = "740104/0020"

        out = self.field.get_encrypted_value(personal_id, self.encryption_key)
        assert_not_equal(personal_id, out)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)
        assert_equal(personal_id, out_decrypt)

    def test_czech_personal_id_smart_field_anonymizer_no_slash(self):
        personal_id = "7401040020"

        out = self.field.get_encrypted_value(personal_id, self.encryption_key)
        assert_not_equal(personal_id, out)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)
        assert_equal(personal_id, out_decrypt)

    def test_czech_personal_id_smart_field_anonymizer_1954_change(self):
        personal_id = "540101/0021"

        out = self.field.get_encrypted_value(personal_id, self.encryption_key)
        assert_not_equal(personal_id, out)

        out_decrypt = self.field.get_decrypted_value(out, self.encryption_key)
        assert_equal(personal_id, out_decrypt)



"""
Implementation of ipcipher.

https://github.com/PowerDNS/ipcipher

IPv4 took from https://github.com/veorq/ipcrypt/blob/939549e3f542c7ae9eb1aec96164f19b5f9fc46c/ipcrypt.py under CC0
"""

# type: ignore
# flake8: noqa
from hashlib import pbkdf2_hmac
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Union

from pyaes import AESModeOfOperationECB


__all__ = ['derive_key', 'encrypt_ipv4', 'decrypt_ipv4', 'encrypt_ipv6', 'decrypt_ipv6', 'encrypt_ip', 'decrypt_ip']

IPv4Type = Union[str, IPv4Address]
IPv6Type = Union[str, IPv6Address]
IPType = Union[IPv4Type, IPv6Type]


def derive_key(key: str):
    """
    PBKDF2(SHA1, Password, 'ipcipheripcipher', 50000, 16)
    """

    return pbkdf2_hmac('sha1', bytes(key, encoding="utf-8"), bytes('ipcipheripcipher', encoding='utf-8'), 50000, 16)


def rotl(b, r):
    return ((b << r) & 0xff) | (b >> (8 - r))


def permute_fwd(state):
    (b0, b1, b2, b3) = state
    b0 += b1
    b2 += b3
    b0 &= 0xff
    b2 &= 0xff
    b1 = rotl(b1, 2)
    b3 = rotl(b3, 5)
    b1 ^= b0
    b3 ^= b2
    b0 = rotl(b0, 4)
    b0 += b3
    b2 += b1
    b0 &= 0xff
    b2 &= 0xff
    b1 = rotl(b1, 3)
    b3 = rotl(b3, 7)
    b1 ^= b2
    b3 ^= b0
    b2 = rotl(b2, 4)
    return b0, b1, b2, b3


def permute_bwd(state):
    (b0, b1, b2, b3) = state
    b2 = rotl(b2, 4)
    b1 ^= b2
    b3 ^= b0
    b1 = rotl(b1, 5)
    b3 = rotl(b3, 1)
    b0 -= b3
    b2 -= b1
    b0 &= 0xff
    b2 &= 0xff
    b0 = rotl(b0, 4)
    b1 ^= b0
    b3 ^= b2
    b1 = rotl(b1, 6)
    b3 = rotl(b3, 3)
    b0 -= b1
    b2 -= b3
    b0 &= 0xff
    b2 &= 0xff
    return b0, b1, b2, b3


def xor4(x, y):
    return [(x[i] ^ y[i]) & 0xff for i in (0, 1, 2, 3)]


def encrypt_ipv4_bytes_key(key: bytes, ip: IPv4Address) -> str:
    """16-byte key, ip string like '192.168.1.2'"""
    if ip.version != 4:
        raise ValueError('IPv6 supplied to IPv4 function.')
    k = [int(x) for x in key]
    try:
        state = [int(x) for x in ip.exploded.split('.')]
    except ValueError:
        raise
    try:
        state = xor4(state, k[:4])
        state = permute_fwd(state)
        state = xor4(state, k[4:8])
        state = permute_fwd(state)
        state = xor4(state, k[8:12])
        state = permute_fwd(state)
        state = xor4(state, k[12:16])
    except IndexError:
        raise
    return '.'.join(str(x) for x in state)


def encrypt_ipv4(key: str, ip: IPv4Type) -> str:
    if not isinstance(ip, IPv4Address):
        ip = ip_address(ip)
    return encrypt_ipv4_bytes_key(derive_key(key), ip)  # type: ignore


def decrypt_ipv4_bytes_key(key: bytes, ip: IPv4Address) -> str:
    """16-byte key, encrypted ip string like '215.51.199.127'"""
    if ip.version != 4:
        raise ValueError('IPv6 supplied to IPv4 function.')
    k = [x for x in key]
    try:
        state = [int(x) for x in ip.exploded.split('.')]
    except ValueError:
        raise
    try:
        state = xor4(state, k[12:16])
        state = permute_bwd(state)
        state = xor4(state, k[8:12])
        state = permute_bwd(state)
        state = xor4(state, k[4:8])
        state = permute_bwd(state)
        state = xor4(state, k[:4])
    except IndexError:
        raise
    return '.'.join(str(x) for x in state)


def decrypt_ipv4(key: str, ip: IPv4Type) -> str:
    if not isinstance(ip, IPv4Address):
        ip = ip_address(ip)
    return decrypt_ipv4_bytes_key(derive_key(key), ip)


def encrypt_ipv6_bytes_key(key: bytes, ip: IPv6Address) -> IPv6Address:
    return IPv6Address(AESModeOfOperationECB(key).encrypt(ip.packed))


def encrypt_ipv6(key: str, ip: IPv6Type) -> str:
    if not isinstance(ip, IPv6Address):
        ip = ip_address(ip)
    return encrypt_ipv6_bytes_key(derive_key(key), ip).compressed


def decrypt_ipv6_bytes_key(key: bytes, ip: IPv6Address) -> IPv6Address:
    return IPv6Address(AESModeOfOperationECB(key).decrypt(ip.packed))


def decrypt_ipv6(key: str, ip: IPv6Type) -> str:
    if not isinstance(ip, IPv6Address):
        ip = ip_address(ip)
    return decrypt_ipv6_bytes_key(derive_key(key), ip).compressed


def encrypt_ip(key: str, ip: IPType) -> str:
    if not isinstance(ip, IPv6Address) and not isinstance(ip, IPv4Address):
        ip = ip_address(ip)
    if ip.version == 4:
        return encrypt_ipv4(key, ip)
    elif ip.version == 6:
        return encrypt_ipv6(key, ip)


def decrypt_ip(key: str, ip: IPType) -> str:
    if not isinstance(ip, IPv6Address) and not isinstance(ip, IPv4Address):
        ip = ip_address(ip)
    if ip.version == 4:
        return decrypt_ipv4(key, ip)
    elif ip.version == 6:
        return decrypt_ipv6(key, ip)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2020-02-17 13:08
from __future__ import unicode_literals

from django.db import migrations
import enumfields.fields
import gdpr.enums


class Migration(migrations.Migration):

    dependencies = [
        ('gdpr', '0006_auto_20190228_0725'),
    ]

    operations = [
        migrations.AddField(
            model_name='legalreason',
            name='state',
            field=enumfields.fields.NumEnumField(default=1, enum=gdpr.enums.LegalReasonState, verbose_name='state'),
        ),
    ]



# Generated by Django 2.1.7 on 2019-03-20 10:06

from django.db import migrations, models
import django.db.models.deletion
import gdpr.mixins


class Migration(migrations.Migration):
    dependencies = [
        ('tests', '0006_auto_20190305_0323'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExtraParentD',
            fields=[
                ('id_d', models.AutoField(editable=False, primary_key=True, serialize=False)),
                ('note', models.CharField(max_length=250)),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='TopParentA',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ParentB',
            fields=[
                ('topparenta_ptr',
                 models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True,
                                      primary_key=True, serialize=False, to='tests.TopParentA')),
                ('birth_date', models.DateField()),
            ],
            options={
                'abstract': False,
            },
            bases=('tests.topparenta',),
        ),
        migrations.CreateModel(
            name='ParentC',
            fields=[
                ('parentb_ptr',
                 models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True,
                                      primary_key=True, serialize=False, to='tests.ParentB')),
                ('first_name', models.CharField(max_length=250)),
            ],
            options={
                'abstract': False,
            },
            bases=('tests.parentb',),
        ),
        migrations.CreateModel(
            name='ChildE',
            fields=[
                ('extraparentd_ptr',
                 models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True,
                                      to='tests.ExtraParentD')),
                ('parentc_ptr',
                 models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True,
                                      primary_key=True, serialize=False, to='tests.ParentC')),
                ('last_name', models.CharField(max_length=250)),
            ],
            options={
                'abstract': False,
            },
            bases=('tests.parentc', 'tests.extraparentd'),
        ),
    ]



"""Minimal django settings to run tests."""
from gdpr.utils import is_reversion_installed

DEBUG = True
SECRET_KEY = 'fake-key'
GDPR_KEY = "GDPR_TEST_KEY_PLEASE_CHANGE-#!6_c+78r-q6)@9@z14=3a754929d600$3ll3)(8h0n@cjc*-CHANGE_ME"
INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "gdpr",
    "tests",
    "django_extensions",
    # Requirements for django-reversion below
    "django.contrib.auth",
    "django.contrib.admin",
    "django.contrib.messages",
    "django.contrib.sessions",
]

if is_reversion_installed():
    INSTALLED_APPS += ["reversion"]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db.sqlite3',
    }
}

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            # insert your TEMPLATE_DIRS here
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                # Insert your TEMPLATE_CONTEXT_PROCESSORS here or use this
                # list if you haven't customized them:
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.request',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

MIDDLEWARE = MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)

USE_TZ = True

# Old legacy settings
# @TODO: Replace
ANONYMIZATION_NAME_KEY = "SOMEFAKEKEY"
ANONYMIZATION_PERSONAL_ID_KEY = 256
ANONYMIZATION_PHONE_KEY = 21212



from typing import Any, Iterable, List, Optional, Union, Type

from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model

from gdpr.encryption import numerize_key
from gdpr.utils import get_number_guess_len, get_reversion_local_field_dict
from gdpr.loading import anonymizer_register


class BaseAnonymizer:
    """
    Base class for Anonymizers defining anonymizer type with properties:
    """


class RelationAnonymizer(BaseAnonymizer):
    """
    Base class for Anonymizers defining special relations.
    """

    model: Type[Model]

    def get_related_objects(self, obj: Model) -> Iterable:
        raise NotImplementedError

    @property
    def model_anonymizer(self):
        return anonymizer_register[self.model]()


class FieldAnonymizer(BaseAnonymizer):
    """
    Field anonymizer's purpose is to anonymize model field according to defined rule.
    """

    ignore_empty_values: bool = True
    empty_values: List[Any] = [None]
    _encryption_key = None
    is_reversible: bool = True

    class IrreversibleAnonymizationException(Exception):
        pass

    def __init__(self, ignore_empty_values: bool = None, empty_values: Optional[List[Any]] = None):
        """
        Args:
            ignore_empty_values: defines if empty value of a model will be ignored or should be anonymized too
            empty_values: defines list of values which are considered as empty
        """
        self._ignore_empty_values = ignore_empty_values if ignore_empty_values is not None else self.ignore_empty_values
        self._empty_values = empty_values if empty_values is not None else self.empty_values

    def get_is_reversible(self, obj=None, raise_exception: bool = False) -> bool:
        """This method allows for custom implementation."""
        if not self.is_reversible and raise_exception:
            raise self.IrreversibleAnonymizationException
        return self.is_reversible

    def get_ignore_empty_values(self, value):
        return self._ignore_empty_values

    def get_is_value_empty(self, value):
        return self.get_ignore_empty_values(value) and value in self._empty_values

    def _get_anonymized_value_from_value(self, value, encryption_key: str):
        if self.get_is_value_empty(value):
            return value
        return self.get_encrypted_value(value, encryption_key)

    def _get_deanonymized_value_from_value(self, obj, value, encryption_key: str):
        if self.get_is_reversible(obj, raise_exception=True):
            if self.get_is_value_empty(value):
                return value
            return self.get_decrypted_value(value, encryption_key)

    def get_value_from_obj(self, obj, name: str, encryption_key: str, anonymization: bool = True):
        if anonymization:
            return self._get_anonymized_value_from_value(getattr(obj, name), encryption_key)
        return self._get_deanonymized_value_from_value(obj, getattr(obj, name), encryption_key)

    def get_value_from_version(self, obj, version, name: str, encryption_key: str, anonymization: bool = True):
        if anonymization:
            return self._get_anonymized_value_from_value(
                get_reversion_local_field_dict(version)[name], encryption_key
            )
        else:
            return self._get_deanonymized_value_from_value(
                obj, get_reversion_local_field_dict(version)[name], encryption_key
            )

    def get_anonymized_value_from_obj(self, obj, name: str, encryption_key: str):
        return self.get_value_from_obj(obj, name, encryption_key, anonymization=True)

    def get_deanonymized_value_from_obj(self, obj, name: str, encryption_key: str):
        return self.get_value_from_obj(obj, name, encryption_key, anonymization=False)

    def get_anonymized_value_from_version(self, obj, version, name: str, encryption_key: str):
        return self.get_value_from_version(obj, version, name, encryption_key, anonymization=True)

    def get_deanonymized_value_from_version(self, obj, version, name: str, encryption_key: str):
        return self.get_value_from_version(obj, version, name, encryption_key, anonymization=False)

    def get_anonymized_value(self, value: Any) -> Any:
        """
        Deprecated
        """
        raise DeprecationWarning()

    def get_encrypted_value(self, value: Any, encryption_key: str) -> Any:
        """
        There must be defined implementation of rule for anonymization

        :param value: value
        :param encryption_key: The encryption key
        :return: Encrypted value
        """
        raise NotImplementedError

    def get_decrypted_value(self, value: Any, encryption_key: str) -> Any:
        """
        There must be defined implementation of rule for deanonymization.

        :param value: Encrypted value
        :param encryption_key: The encryption key
        :return: Decrypted value
        """
        if self.get_is_reversible(raise_exception=True):
            raise NotImplementedError


class NumericFieldAnonymizer(FieldAnonymizer):
    max_anonymization_range: Optional[int] = None

    def __init__(self, max_anonymization_range: int = None, ignore_empty_values: bool = None,
                 empty_values: Optional[List[Any]] = None):
        if max_anonymization_range is not None:
            self.max_anonymization_range = max_anonymization_range
        super().__init__(ignore_empty_values, empty_values)

    def get_numeric_encryption_key(self, encryption_key: str, value: Union[int, float] = None) -> int:
        """
        From `encryption_key` create it's numeric counterpart of appropriate length.

        If value is supplied then the appropriate length is based on it if not the
        parameter `self.max_anonymization_range` is used.

        If the numeric_encryption_key is used the value+key may be one order bigger then the original value.

        e.g. value=5, numeric_encryption_key=8 => len('13') == 2

        :param encryption_key: The encryption key generated by anonymizer.
        :param value: Value to which the result of this function will be used.
        :return: Numeric counterpart of encryption_key
        """
        if value is None:
            if self.max_anonymization_range is None:
                return numerize_key(encryption_key)
            return numerize_key(encryption_key) % self.max_anonymization_range

        return numerize_key(encryption_key) % 10 ** get_number_guess_len(value)



"""Since django 1.11 djnago-GIS requires GDAL."""
import logging
from typing import Optional

from django.core.exceptions import ImproperlyConfigured

from gdpr.anonymizers.base import NumericFieldAnonymizer

logger = logging.getLogger(__name__)


def is_gis_installed():
    try:
        from django.contrib.gis.geos import Point
        return True
    except ImproperlyConfigured:
        return False


if not is_gis_installed():
    logger.warning('Unable to load django GIS. GIS anonymization disabled.')


class ExperimentalGISPointFieldAnonymizer(NumericFieldAnonymizer):
    """
    Anonymizer for PointField from django-gis.

    Warnings:
        May not fully work. Currently works only on positive coordinates.
        With ``max_x_range`` and ``max_y_range`` specified. Also anonymization occurs only on the whole part.
    """

    max_x_range: int
    max_y_range: int

    def __init__(self, max_x_range: Optional[int] = None, max_y_range: Optional[int] = None, *args, **kwargs):
        if max_x_range is not None:
            self.max_x_range = max_x_range
        elif self.max_x_range is None:
            raise ImproperlyConfigured(f'{self.__class__} does not have `max_x_range`.')
        if max_y_range is not None:
            self.max_y_range = max_y_range
        elif self.max_y_range is None:
            raise ImproperlyConfigured(f'{self.__class__} does not have `max_y_range`.')
        super().__init__(*args, **kwargs)

    def get_encrypted_value(self, value, encryption_key: str):
        if not is_gis_installed():
            raise ImproperlyConfigured('Unable to load django GIS.')
        from django.contrib.gis.geos import Point

        new_val: Point = Point(value.tuple)
        new_val.x = (new_val.x + self.get_numeric_encryption_key(encryption_key, int(new_val.x))) % self.max_x_range
        new_val.y = (new_val.y + self.get_numeric_encryption_key(encryption_key, int(new_val.y))) % self.max_y_range

        return new_val

    def get_decrypted_value(self, value, encryption_key: str):
        if not is_gis_installed():
            raise ImproperlyConfigured('Unable to load django GIS.')
        from django.contrib.gis.geos import Point

        new_val: Point = Point(value.tuple)
        new_val.x = (new_val.x - self.get_numeric_encryption_key(encryption_key, int(new_val.x))) % self.max_x_range
        new_val.y = (new_val.y - self.get_numeric_encryption_key(encryption_key, int(new_val.y))) % self.max_y_range

        return new_val



# Generated by Django 3.1 on 2021-04-16 11:47

from django.db import migrations, models
import gdpr.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0007_childe_extraparentd_parentb_parentc_topparenta'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomerRegistration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email_address', models.EmailField(blank=True, max_length=254, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
    ]



from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Type, Union

from django.db.models import Model

from gdpr.loading import anonymizer_register

FieldMatrix = Union[str, Tuple[Any, ...]]
FieldList = Union[List[str], str]
RelatedFieldDict = Dict[str, "Fields"]

if TYPE_CHECKING:
    from gdpr.anonymizers import ModelAnonymizer


class Fields:
    local_fields: FieldList
    related_fields: RelatedFieldDict
    anonymizer: "ModelAnonymizer"
    model: Type[Model]

    def __init__(self, fields: FieldMatrix, model: Type[Model], anonymizer_instance: "ModelAnonymizer" = None):
        self.model = model
        self.anonymizer = anonymizer_register[self.model]() if anonymizer_instance is None else anonymizer_instance
        self.local_fields = self.parse_local_fields(fields)
        self.related_fields = self.parse_related_fields(fields)

    def parse_local_fields(self, fields: FieldMatrix) -> FieldList:
        """Get Iterable of local fields from fields matrix."""
        if fields == '__ALL__' or ('__ALL__' in fields and type(fields) != str):
            return list(self.anonymizer.keys())

        return [field for field in fields if type(field) == str]

    def parse_related_fields(self, fields: FieldMatrix) -> RelatedFieldDict:
        """Get Dictionary of related fields from fields matrix."""
        out_dict = {}
        for name, related_fields in [field_tuple for field_tuple in fields if isinstance(field_tuple, (list, tuple))]:
            out_dict[name] = Fields(
                related_fields,
                self.anonymizer.get_related_model(name),
                anonymizer_instance=self.anonymizer.get_related_model_anonymizer_none(name)
            )

        return out_dict

    def get_tuple(self) -> FieldMatrix:
        return (*self.local_fields, *[(name, fields.get_tuple()) for name, fields in self.related_fields.items()])

    def __len__(self):
        return len(self.local_fields) + len(self.related_fields)

    def __isub__(self, other: "Fields") -> "Fields":
        self.local_fields = [field for field in self.local_fields if field not in other.local_fields]

        for name, related_fields in self.related_fields.items():
            if name in other.related_fields:
                related_fields -= other.related_fields[name]

        for name in list(self.related_fields.keys()):
            if len(self.related_fields[name]) == 0:
                del self.related_fields[name]

        return self



from django.utils.translation import gettext_lazy as _

from enumfields import Choice, ChoiceEnum


class LegalReasonState(ChoiceEnum):

    ACTIVE = Choice(1, _('Active'))
    EXPIRED = Choice(2, _('Expired'))
    DEACTIVATED = Choice(3, _('Deactivated'))



from unittest import skipIf

from django.test import TestCase

from gdpr.anonymizers.gis import ExperimentalGISPointFieldAnonymizer, is_gis_installed
from germanium.tools import assert_not_equal, assert_tuple_equal


class TestGISPointFieldAnonymizer(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.field = ExperimentalGISPointFieldAnonymizer(max_x_range=100, max_y_range=100)
        cls.encryption_key = 'LoremIpsumDolorSitAmet'

    @skipIf(not is_gis_installed(), 'Django GIS not available.')
    def test_point_base(self):
        from django.contrib.gis.geos import Point

        point = Point(1, 1)
        out = self.field.get_encrypted_value(point, self.encryption_key)

        assert_not_equal(point.tuple, out.tuple)

        out_decrypted = self.field.get_decrypted_value(out, self.encryption_key)

        assert_tuple_equal(point.tuple, out_decrypted.tuple)



import datetime
import re
from typing import Any, Optional, Tuple, Union

from django.core.exceptions import ValidationError

from gdpr.anonymizers.base import FieldAnonymizer, NumericFieldAnonymizer
from gdpr.encryption import LETTERS_UPPER, NUMBERS, decrypt_text, encrypt_text

PRE_NUM_WEIGHTS = [10, 5, 8, 4, 2, 1]
NUM_WEIGHTS = [6, 3, 7, 9, 10, 5, 8, 4, 2, 1]


class CzechAccountNumber:
    pre_num: Optional[int]
    pre_num_len: Optional[int]
    num: int
    num_len: int = 10
    bank: int

    CZECH_ACCOUNT_RE = re.compile('((?P<pre_num>[0-9]{0,6})-)?(?P<num>[0-9]{1,10})/(?P<bank_code>[0-9]{4})')

    def __init__(self, num: Union[int, str], bank: Union[int, str], pre_num: Optional[Union[int, str]] = None,
                 num_len: int = 10, pre_num_len: Optional[int] = None, bank_len: int = 4):
        self.num = int(num)
        self.num_len = num_len
        self.bank = int(bank)
        self.bank_len = bank_len
        self.pre_num_len = pre_num_len
        self.pre_num = int(pre_num) if pre_num is not None and pre_num != 0 else None

    def check_account_format(self) -> bool:
        pre_num = "%06d" % (self.pre_num or 0)
        num = '0' * (10 - len(str(self.num))) + str(self.num)

        pre_num_valid = sum(map(lambda x: x[0] * x[1], zip(map(lambda x: int(x), pre_num), PRE_NUM_WEIGHTS))) % 11 == 0
        num_valid = sum(map(lambda x: x[0] * x[1], zip(map(lambda x: int(x), num), NUM_WEIGHTS))) % 11 == 0

        return num_valid and pre_num_valid

    def _brute_force_next(self):
        self.num += 1
        if len(str(self.num)) > 10:
            self.num = 0
        while not self.check_account_format():
            if len(str(self.num)) > 10:
                self.num = 0
            self.num += 1

    def brute_force_next(self, n: int):
        for i in range(n):
            self._brute_force_next()

        return self  # allow chaining

    def _brute_force_prev(self):
        self.num -= 1
        if self.num <= 0:
            self.num = int('9' * 10)
        while not self.check_account_format():
            if self.num <= 0:
                self.num = int('9' * 10)
            self.num -= 1

    def brute_force_prev(self, n: int):
        for i in range(n):
            self._brute_force_prev()

        return self  # allow chaining

    @classmethod
    def parse(cls, value: str) -> "CzechAccountNumber":
        """
        :param value:
        :return: AccountNumber(predcisli)-(cislo)/(kod_banky)
        """
        account = re.match(cls.CZECH_ACCOUNT_RE, value)

        if account is None:
            raise ValidationError(f'Str \'{value}\' does not appear to be czech account number.')

        pre_num = account.group('pre_num')
        num = account.group('num')
        bank_code = account.group('bank_code')
        return cls(pre_num=pre_num, pre_num_len=len(pre_num or ""), num=num, num_len=len(num),
                   bank=bank_code, bank_len=len(bank_code))

    def __str__(self):
        return ((f'{str(self.pre_num).rjust(self.pre_num_len, "0") if self.pre_num_len else self.pre_num}-'
                 if self.pre_num else ''
                 ) + f'{str(self.num).rjust(self.num_len, "0")}/{str(self.bank).rjust(self.bank_len, "0")}')


class CzechIBAN(CzechAccountNumber):
    has_spaces = False
    control_code: int
    CZECH_IBAN_RE = re.compile(
        'CZ(?P<control_code>[0-9]{2}) ?(?P<bank_code>[0-9]{4}) ?'
        '(?P<pre_num>[0-9]{4} ?[0-9]{2})(?P<num>[0-9]{2} ?[0-9]{4} ?[0-9]{4})',
    )

    def __init__(self, *args, has_spaces: bool = False, control_code: Optional[int] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.has_spaces = has_spaces
        if control_code is not None:
            self.control_code = control_code
        else:
            self._update_control_code()

    @classmethod
    def from_account(cls, value: "CzechAccountNumber") -> "CzechIBAN":
        return cls(num=value.num, bank=value.bank, pre_num=value.num)

    @classmethod
    def parse(cls, value: str) -> "CzechIBAN":
        """
        :param value:
        :return: AccountNumber(predcisli)-(cislo)/(kod_banky)
        """
        account = re.match(cls.CZECH_IBAN_RE, value)

        if account:
            control_code = account.group('control_code').upper()
            bank_code = account.group('bank_code').upper()
            pre_num = account.group('pre_num').replace(' ', '').upper()
            num = account.group('num').replace(' ', '').upper()

            if all([all([i in (LETTERS_UPPER + NUMBERS) for i in control_code]),
                    all([i in NUMBERS for i in bank_code]),
                    all([i in NUMBERS for i in pre_num]),
                    all([i in NUMBERS for i in num])]):
                return cls(
                    control_code=int(control_code), has_spaces=' ' in value,
                    pre_num=int(pre_num), num=int(num), bank=int(bank_code))

        raise ValidationError(f'IBAN \'{value}\' does not appear to be czech IBAN.')

    def _to_str(self, spaces: Optional[bool] = None):
        pre_num = str(self.pre_num or 0).rjust(6, '0')
        num = str(self.num).rjust(10, '0')
        out = (f'CZ{str(self.control_code).rjust(2, "0")} {str(self.bank).rjust(4, "0")} {pre_num[:4]} '
               f'{pre_num[4:]}{num[:2]} {num[2:6]} {num[6:]}')
        if (spaces is None and self.has_spaces) or spaces:
            return out
        return out.replace(' ', '')

    def __str__(self) -> str:
        return self._to_str()

    def _to_int(self) -> int:
        """
        Convert IBAN to numeric value (ISO 7064) to be able to calculate `control_code`and check if IBAN is valid.
        """
        iban = self._to_str(spaces=False)
        return int(
            "".join([i if not ord('A') <= ord(i) <= ord('Z') else str(ord(i) - 55) for i in iban[4:] + iban[:4]]))

    def check_iban_format(self) -> bool:
        return self._to_int() % 97 == 1

    def _update_control_code(self):
        """
        Calculate new control code
        """
        self.control_code = 0
        self.control_code = 98 - (self._to_int() % 97)

    def brute_force_next(self, n: int):
        self.control_code = 0
        super().brute_force_next(n)
        self._update_control_code()

        return self  # allow chaining

    def brute_force_prev(self, n: int):
        self.control_code = 0
        super().brute_force_prev(n)
        self._update_control_code()

        return self  # allow chaining


class CzechPersonalID:
    date: datetime.date
    day_index: int
    control_number: Optional[int] = None

    day_offset: bool = False
    is_male: bool
    is_extra: bool = False
    has_slash: bool = True

    CZECH_PERSONAL_ID_RE = re.compile(
        r'^(?P<year>\d{2})(?P<month>\d{2})(?P<day>\d{2})/?(?P<day_index>\d{3})(?P<key>\d)?$')

    def __init__(self, date: datetime.date, is_male: bool, day_index: int, control_number: Optional[int] = None,
                 is_extra: bool = False, day_offset: bool = False, has_slash: bool = True):
        """

        Args:
            date: The ``datetime.date`` object of the Personal ID
            is_male: Is the person male? (Month + 50)
            day_index: The index of the person in the given day
            control_number: The last digit in personal id
            is_extra: Did we run out of IDs on that day? (Month + 20)
            day_offset: In some rare cases the days may be offset by 50 is this the case?
            has_slash: Show ``/`` in ``__str__`` representation

        """
        self.day_index = day_index
        self.date = date
        self.is_male = is_male
        self.control_number = control_number
        self.is_extra = is_extra
        self.day_offset = day_offset
        self.has_slash = has_slash

    @property
    def is_pre_1954(self):
        return self.date.year < 1954

    def __str__(self):
        month = self.date.month
        if not self.is_male:
            month += 50
        if self.is_extra:
            month += 20
        return (f'{str(self.date.year)[-2:]}{"%02d" % month}'
                f'{"%02d" % (self.date.day if not self.day_offset else self.date.day + 50)}'
                f'{"/" if self.has_slash else ""}{("%03d" % self.day_index)}'
                f'{self.control_number if self.control_number is not None else ""}')

    @classmethod
    def parse(cls, value) -> "CzechPersonalID":
        personal_id = re.match(cls.CZECH_PERSONAL_ID_RE, value)

        if personal_id is None:
            raise ValidationError(f'Str \'{value}\' does not appear to be czech personal id.')

        year = int(personal_id.group('year'))
        month = int(personal_id.group('month'))
        day = int(personal_id.group('day'))
        day_index = int(personal_id.group('day_index'))
        key = int(personal_id.group('key')) if personal_id.group('key') is not None else None

        pre_1954 = len(value.replace('/', '')) == 9
        is_male = month < 50
        is_extra = month > 12 if is_male else month > 62
        is_day_offset = day > 50
        full_year = (2000 if year < 54 and not pre_1954 else 1900) + year
        if is_male and is_extra and full_year > 2003:
            month -= 20
        elif not is_male and is_extra and full_year > 2003:
            month -= 70
        elif not is_male:
            month -= 50
        if not (1 <= month <= 12):
            raise ValidationError(f'Str \'{value}\' does not appear to be czech personal id.')

        if is_day_offset:
            day -= 50

        return cls(
            date=datetime.date(year=full_year, month=month, day=day),
            is_male=is_male,
            day_index=day_index,
            control_number=key,
            is_extra=is_extra,
            day_offset=is_day_offset,
            has_slash='/' in value,
        )

    def check_format(self):

        # Three digits for verification number were used until 1. january 1954
        if not self.is_pre_1954:
            """
            Fourth digit has been added since 1. January 1954.
            It is modulo of dividing birth number and verification number by 11.
            If the modulo were 10, the last number was 0 (and therefore, the whole
            birth number weren't dividable by 11. These number are no longer used (since 1985)
            and condition 'modulo == 10' can be removed some years after 2085.
            """

            modulo = int(str(self).replace('/', '')[:-1]) % 11

            if (modulo != self.control_number) and (modulo != 10 or self.control_number != 0):
                return False

        return True

    def brute_force_control_number(self):
        self.control_number = 0
        while not self.check_format():
            self.control_number += 1

    def encrypt(self, numeric_key):
        numeric_key %= 365

        if numeric_key % 2 == 0:
            self.is_male = not self.is_male

        self.date -= datetime.timedelta(days=numeric_key + 1)
        self.day_index = int(encrypt_text(str(numeric_key), str(self.day_index), NUMBERS))

        if self.is_pre_1954:
            self.control_number = None
        else:
            self.brute_force_control_number()

        return self  # Enable chaining

    def decrypt(self, numeric_key):
        numeric_key %= 365

        if numeric_key % 2 == 0:
            self.is_male = not self.is_male

        self.date += datetime.timedelta(days=numeric_key + 1)
        self.day_index = int(decrypt_text(str(numeric_key), str(self.day_index), NUMBERS))

        if self.is_pre_1954:
            self.control_number = None
        else:
            self.brute_force_control_number()

        return self  # Enable chaining


class CzechAccountNumberFieldAnonymizer(NumericFieldAnonymizer):
    """
    Anonymization for czech account number.

    Setting `use_smart_method=True` retains valid format for encrypted value using this
    have significant effect on performance.
    """
    use_smart_method = False
    max_anonymization_range = 10000

    def __init__(self, *args, use_smart_method=False, **kwargs):
        self.use_smart_method = use_smart_method
        super().__init__(*args, **kwargs)

    def get_encrypted_value(self, value, encryption_key: str):
        account = CzechAccountNumber.parse(value)

        if self.use_smart_method and account.check_account_format():
            return str(account.brute_force_next(self.get_numeric_encryption_key(encryption_key)))

        account.num = int(encrypt_text(encryption_key, str(account.num), NUMBERS))

        return str(account)

    def get_decrypted_value(self, value: Any, encryption_key: str):
        account = CzechAccountNumber.parse(value)

        if self.use_smart_method and account.check_account_format():
            return str(account.brute_force_prev(self.get_numeric_encryption_key(encryption_key)))

        account.num = int(decrypt_text(encryption_key, str(account.num), NUMBERS))

        return str(account)


class CzechIBANSmartFieldAnonymizer(NumericFieldAnonymizer):
    max_anonymization_range = 10000

    def get_encrypted_value(self, value: Any, encryption_key: str):
        iban = CzechIBAN.parse(value)
        if not iban.check_iban_format():
            raise ValidationError(f'IBAN \'{value}\' does not appear to be valid czech IBAN.')
        return str(iban.brute_force_next(self.get_numeric_encryption_key(encryption_key)))

    def get_decrypted_value(self, value: Any, encryption_key: str):
        iban = CzechIBAN.parse(value)
        if not iban.check_iban_format():
            raise ValidationError(f'IBAN \'{value}\' does not appear to be valid czech IBAN.')
        return str(iban.brute_force_prev(self.get_numeric_encryption_key(encryption_key)))


class CzechPhoneNumberFieldAnonymizer(FieldAnonymizer):

    def split_phone_number(self, value: str) -> Tuple[str, str]:
        """
        Split phone number in international format into area code and number
        """
        area_code = value[:-9]
        phone_number = value[-9:]
        return area_code, phone_number

    def get_encrypted_value(self, value: str, encryption_key: str):
        area_code, phone_number = self.split_phone_number(value)
        encrypted_phone_number = encrypt_text(encryption_key, phone_number[3:], NUMBERS)
        return f'{area_code}{phone_number[:3]}{encrypted_phone_number}'

    def get_decrypted_value(self, value: str, encryption_key: str):
        area_code, phone_number = self.split_phone_number(value)
        encrypted_phone_number = decrypt_text(encryption_key, phone_number[3:], NUMBERS)
        return f'{area_code}{phone_number[:3]}{encrypted_phone_number}'


class CzechIDCardFieldAnonymizer(NumericFieldAnonymizer):
    max_anonymization_range = int("9" * 9)

    def get_encrypted_value(self, value: str, encryption_key: str):
        return f"{value[0]}{encrypt_text(str(self.get_numeric_encryption_key(encryption_key)), value[1:], NUMBERS)}"

    def get_decrypted_value(self, value: str, encryption_key: str):
        return f"{value[0]}{decrypt_text(str(self.get_numeric_encryption_key(encryption_key)), value[1:], NUMBERS)}"


class CzechPersonalIDSmartFieldAnonymizer(NumericFieldAnonymizer):
    max_anonymization_range = 365

    def get_encrypted_value(self, value: str, encryption_key: str):
        return str(CzechPersonalID.parse(value).encrypt(self.get_numeric_encryption_key(encryption_key)))

    def get_decrypted_value(self, value: str, encryption_key: str):
        return str(CzechPersonalID.parse(value).decrypt(self.get_numeric_encryption_key(encryption_key)))



# Generated by Django 2.1.5 on 2019-02-08 17:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('tests', '0002_note'),
    ]

    operations = [
        migrations.AddField(
            model_name='account',
            name='IBAN',
            field=models.CharField(blank=True, max_length=34, null=True),
        ),
        migrations.AddField(
            model_name='account',
            name='swift',
            field=models.CharField(blank=True, max_length=11, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-04-16 10:07
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='AnonymizedData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('field', models.CharField(max_length=250, verbose_name='anonymized field name')),
                ('object_id', models.TextField(verbose_name='related object ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType',
                                   verbose_name='related object content type')),
            ],
            options={
                'verbose_name': 'anonymized data',
                'verbose_name_plural': 'anonymized data',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='LegalReason',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('expires_at', models.DateTimeField(db_index=True, verbose_name='expires at')),
                ('tag', models.CharField(blank=True, max_length=100, null=True, verbose_name='tag')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('purpose_slug', models.CharField(max_length=100, verbose_name='purpose')),
                ('source_object_id', models.TextField(verbose_name='source object ID')),
                ('source_object_content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType',
                                   verbose_name='source object content type')),
            ],
            options={
                'verbose_name': 'legal reason',
                'verbose_name_plural': 'legal reasons',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='LegalReasonRelatedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_id', models.TextField(verbose_name='related object ID')),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType',
                                   verbose_name='related object content type')),
                ('legal_reason',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='related_objects',
                                   to='gdpr.LegalReason', verbose_name='legal reason')),
            ],
            options={
                'verbose_name': 'legal reason related object',
                'verbose_name_plural': 'legal reasons related objects',
                'ordering': ('-created_at',),
            },
        ),
        migrations.AddField(
            model_name='anonymizeddata',
            name='expired_reason',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE,
                                    to='gdpr.LegalReason', verbose_name='expired reason'),
        ),
    ]



from typing import Any, List, Type

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Model, QuerySet


def str_to_class(class_string: str) -> Any:
    module_name, class_name = class_string.rsplit('.', 1)
    # load the module, will raise ImportError if module cannot be loaded
    m = __import__(module_name, globals(), locals(), [str(class_name)])
    # get the class, will raise AttributeError if class cannot be found
    c = getattr(m, class_name)
    return c


def get_number_guess_len(value):
    """
    Safety measure against key getting one bigger (overflow) on decrypt e.g. (5)=1 -> 5 + 8 = 13 -> (13)=2
    Args:
        value: Number convertible to int to get it's length

    Returns:
        The even length of the whole part of the number
    """
    guess_len = len(str(int(value)))
    return guess_len if guess_len % 2 != 0 else (guess_len - 1)


def get_field_or_none(model: Type[Model], field_name: str):
    """
    Use django's _meta field api to get field or return None.

    Args:
        model: The model to get the field on
        field_name: The name of the field

    Returns:
        The field or None

    """
    try:
        return model._meta.get_field(field_name)
    except FieldDoesNotExist:
        return None


"""
Enable support for druids reversion fork
"""


def get_reversion_versions(obj: Any) -> QuerySet:
    from reversion.models import Version
    from django.contrib.contenttypes.models import ContentType

    return Version.objects.get_for_object(obj)


def get_reversion_version_model(version) -> Type[Model]:
    """Get object model of the version."""
    if hasattr(version, '_model'):
        return version._model
    return version.content_type.model_class()


def get_reversion_local_field_dict(obj):
    if hasattr(obj, '_local_field_dict'):
        return obj._local_field_dict
    return obj.flat_field_dict


def is_reversion_installed():
    try:
        import reversion
        return True
    except ImportError:
        return False


def get_all_parent_objects(obj: Model) -> List[Model]:
    """Return all model parent instances."""
    parent_paths = [
        [path_info.join_field.name for path_info in parent_path]
        for parent_path in
        [obj._meta.get_path_to_parent(parent_model) for parent_model in obj._meta.get_parent_list()]
    ]

    parent_objects = []
    for parent_path in parent_paths:
        parent_obj = obj
        for path in parent_path:
            parent_obj = getattr(parent_obj, path, None)
        parent_objects.append(parent_obj)

    return [i for i in parent_objects if i is not None]


def get_all_obj_and_parent_versions_queryset_list(obj: Model) -> List[QuerySet]:
    """Return list of object and its parent version querysets"""
    from gdpr.utils import get_reversion_versions

    return [get_reversion_versions(i) for i in get_all_parent_objects(obj)] + [get_reversion_versions(obj)]


def get_all_obj_and_parent_versions(obj: Model) -> List[Model]:
    """Return list of all object and its parent versions"""
    return [item for sublist in get_all_obj_and_parent_versions_queryset_list(obj) for item in sublist]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2019-02-27 14:44
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import gdpr.mixins
import tests.validators


class Migration(migrations.Migration):
    dependencies = [
        ('tests', '0004_facebook_id'),
    ]

    operations = [
        migrations.CreateModel(
            name='Avatar',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.FileField(upload_to='')),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
        migrations.AlterField(
            model_name='account',
            name='number',
            field=models.CharField(blank=True, max_length=256, null=True,
                                   validators=[tests.validators.BankAccountValidator]),
        ),
        migrations.AlterField(
            model_name='customer',
            name='personal_id',
            field=models.CharField(blank=True, max_length=10, null=True,
                                   validators=[tests.validators.CZBirthNumberValidator]),
        ),
        migrations.AddField(
            model_name='avatar',
            name='custormer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='avatars',
                                    to='tests.Customer'),
        ),
    ]



from collections import OrderedDict, _OrderedDictItemsView, _OrderedDictKeysView, _OrderedDictValuesView
from importlib import import_module
from typing import TYPE_CHECKING, Any, Generic, Iterator, Optional, Type, TypeVar, Union

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from django.utils.encoding import force_text

from .utils import str_to_class

if TYPE_CHECKING:
    from gdpr.anonymizers import ModelAnonymizer
    from gdpr.purposes import AbstractPurpose


class BaseLoader:
    """Base class for all loaders."""

    def import_modules(self) -> None:
        raise NotImplementedError


class AppLoader(BaseLoader):
    """Scan all installed apps for `module_name` module."""

    module_name: str

    def import_modules(self) -> None:
        for app in apps.get_app_configs():
            if app.name == 'gdpr':
                continue
            try:
                import_module(f'{app.name}.{self.module_name}')
            except ImportError as ex:
                if force_text(ex) != f'No module named \'{app.name}.{self.module_name}\'':
                    raise ex


class SettingsListLoader(BaseLoader):
    """Import all modules from list `list_name` in settings."""

    list_name: str

    def import_modules(self):
        if not hasattr(settings, self.list_name):
            raise ImproperlyConfigured(f'settings.{self.list_name} not found.')
        modules_list = getattr(settings, self.list_name)
        if type(modules_list) in [list, tuple]:
            for i in modules_list:
                import_module(i)
        else:
            raise ImproperlyConfigured(f'settings.{self.list_name} have incorrect type {str(type(modules_list))}.')


class SettingsListAnonymizerLoader(SettingsListLoader):
    """Load all anonymizers from settings.GDPR_ANONYMIZERS_LIST list."""

    list_name = 'GDPR_ANONYMIZERS_LIST'


class SettingsListPurposesLoader(SettingsListLoader):
    """Load all purposes from settings.GDPR_PURPOSES_LIST list."""

    list_name = 'GDPR_PURPOSES_LIST'


class AppAnonymizerLoader(AppLoader):
    """Scan all installed apps for anonymizers module which should contain anonymizers."""

    module_name = 'anonymizers'


class AppPurposesLoader(AppLoader):
    """Scan all installed apps for purposes module which should contain purposes."""

    module_name = 'purposes'


K = TypeVar('K')
V = TypeVar('V')


class BaseRegister(Generic[K, V]):
    """Base class for all registers."""

    _is_import_done = False
    register_dict: "OrderedDict[K, V]"
    loaders_settings: str
    default_loader: Optional[str]

    def __init__(self):
        self.register_dict = OrderedDict()

    def register(self, key: K, object_class: V) -> None:
        self.register_dict[key] = object_class

    def _import_objects(self) -> None:
        default_loader = [self.default_loader] if self.default_loader else []
        for loader_path in getattr(settings, self.loaders_settings, default_loader):
            if isinstance(loader_path, (list, tuple)):
                for path in loader_path:
                    import_module(path)
            else:
                str_to_class(loader_path)().import_modules()

    def _import_objects_once(self) -> None:
        if self._is_import_done:
            return
        self._is_import_done = True
        self._import_objects()

    def __iter__(self) -> Iterator[V]:
        self._import_objects_once()

        for o in self.register_dict.values():
            yield o

    def __contains__(self, key: K) -> bool:
        self._import_objects_once()

        return key in self.register_dict.keys()

    def __getitem__(self, key: K) -> V:
        self._import_objects_once()

        return self.register_dict[key]

    def keys(self) -> "_OrderedDictKeysView[K]":
        self._import_objects_once()

        return self.register_dict.keys()

    def items(self) -> "_OrderedDictItemsView[K, V]":
        self._import_objects_once()

        return self.register_dict.items()

    def values(self) -> "_OrderedDictValuesView[V]":
        self._import_objects_once()

        return self.register_dict.values()

    def get(self, *args, **kwargs) -> Union[V, Any]:
        self._import_objects_once()

        return self.register_dict.get(*args, **kwargs)


class AnonymizersRegister(BaseRegister[Model, Type["ModelAnonymizer"]]):
    """
    AnonymizersRegister is storage for found anonymizer classes.
    """

    default_loader = 'gdpr.loading.AppAnonymizerLoader'
    loaders_settings = 'ANONYMIZATION_LOADERS'


class PurposesRegister(BaseRegister[str, Type["AbstractPurpose"]]):
    """
    PurposesRegister is storage for found purpose classes.
    """

    default_loader = 'gdpr.loading.AppPurposesLoader'
    loaders_settings = 'PURPOSE_LOADERS'


anonymizer_register = AnonymizersRegister()
purpose_register = PurposesRegister()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2019-02-25 16:27
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('tests', '0003_IBAN'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='customer',
            name='fb_id',
        ),
        migrations.AddField(
            model_name='customer',
            name='facebook_id',
            field=models.CharField(blank=True, help_text='Facebook ID used for login via Facebook.', max_length=256,
                                   null=True, verbose_name='Facebook ID'),
        ),
    ]
